#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Exceptions raised by PyScribus.
"""

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#


class InsaneSLAValue(Exception):
    """
    Exception raised if a XML value or a set of XML values is/are
    not possible, are both in contradiction.

    This usually happens if someone edited a SLA file and wroted some
    wrong values in it.

    For example, a list can't be a numeroted and bullet one, Scribus UI
    doesn't allow that but manual SLA editing does.
    """


class MissingSLAAttribute(Exception):
    """
    Exception raised when a mandatory XML attribute in SLA file is missing.

    An exception more precise than InsaneSLAValue.
    """


class ConflictingLayer(Exception):
    """
    Exception raised if a document Layer have same attributes that an other
    one.
    """


# --- Items attributes -----------------------------------------


class UnknownOrEmptyItemAttributeType(Exception):
    """
    Exception raised when a ``itemattribute.ItemAttribute`` (and inherited
    classes) has a unknown or empty type, as it is not in
    ``itemattribute.ItemAttribute.attrib_types``
    """


# --- Quick setup ----------------------------------------------


class QuickSetupInvalidValue(Exception):
    """
    Exception raised when a quick setup value is invalid.
    """


# --- Unknown SLA references -----------------------------------


class UnknownLayer(Exception):
    """
    Exception raised if a SLA element points to a layer unknown to the
    document.
    """


class UnknownStyleInStory(Exception):
    """
    Exception raised if a SLA element points to a style unknown to the
    document.
    """


class InvalidColor(Exception):
    """
    Exception raised when color inks (CMYK or RGB) are invalid, incomplete
    or doesn't even exists.

    - RGB inks must range from 0 to 255.
    - CMYK inks must range from 0 to 100.
    """


class ColorMixing(Exception):
    """
    Exception raised when two colors do not share the same color space (CMYK
    or RVB).
    """


class UnknownRenderBufferProperty(Exception):
    """
    Exception raised when a render buffer property is missing.
    """


# --- dimensions -----------------------------------------------


class UnknownDimUnit(Exception):
    """
    That ``dimensions.Dim`` unit is not handled.
    """


class InvalidDim(Exception):
    """
    Exception raised if a ``dimensions.Dim`` value is impossible according
    to its unit.

    - Pica points must no be inferior to 0.
    - (Ordinary) angle (deg unit) must range from 0 to 180.
    - Calligraphic pen angle (cdeg unit) must range from 0 to 180.
    """


class IncompatibleDim(Exception):
    """
    Exception raised on incompatible ``dimensions.Dim`` units conversions.
    """


# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus

Scribus file. Document classes.
"""

# Imports ===============================================================#

# Standard library ---------------------------------------------

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

import copy

from typing import Any, Union, Optional, Literal, NoReturn
from collections import OrderedDict

# PyScribus model ----------------------------------------------

import pyscribus.model.layers as layers
import pyscribus.model.dimensions as dimensions
import pyscribus.model.pageobjects as pageobjects

from pyscribus.model.sla import SLA
from pyscribus.model.pages import Page, MasterPage

# PyScribus file -----------------------------------------------

import pyscribus.file as pyf
import pyscribus.file.exceptions as pyfe

# Global variables / Annotations ========================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

FirstLast = Literal["first", "last"]
StringOrNone = Union[str, None]

# Classes ===============================================================#

class Metadatas:
    """
    Interface for document's metadatas.

    :Example:

    .. code:: python

       from pyscribus.file import ScribusFile

       doc = ScribusFile("1.5.5", "example.sla")

       title = doc.metadatas.title
       date = doc.metadatas.date

       doc.metadatas.add_keyword("NewKeyword")

       metadatas_dump = dict(doc.metadatas)

    """

    def __init__(self, model: SLA):
        self.model = model

    @pyfe.has_document
    def __get(self, key) -> Union[str, list[str], None]:
        return self.model.document.metadata.get(key)

    def __iter__(self):
        """
        Special method to make Metadatas class iterable.
        """
        for key, value in self.model.document.metadata.items():
            yield (key, value)

    # Keywords -----------------------------------------------------------

    @property
    def keywords(self) -> Union[list[str], None]:
        return self.__get("keywords")

    def add_keyword(self, new_keyword: str) -> NoReturn:
        """
        Add a new keyword to the document's metadatas.

        :type new_keyword: str
        :param new_keyword: New keyword to add.
        """

        keywords = self.__get("keywords")

        if keywords is None:
            keywords = []
        else:
            keywords = list(set(keywords))

        if new_keyword not in keywords:
            keywords.append(new_keyword)

        self.model.document.metadata["keywords"] = keywords

    # --------------------------------------------------------------------

    @property
    def title(self) -> StringOrNone:
        return self.__get("title")

    @property
    def author(self) -> StringOrNone:
        """Author(s) of the document"""
        return self.__get("author")

    @property
    def subject(self) -> StringOrNone:
        return self.__get("subject")

    @property
    def comments(self) -> StringOrNone:
        return self.__get("comments")

    @property
    def publisher(self) -> StringOrNone:
        return self.__get("publisher")

    @property
    def contributor(self) -> StringOrNone:
        """Contributor(s) to the document"""
        return self.__get("contributor")

    @property
    def date(self) -> StringOrNone:
        return self.__get("date")

    @property
    def type(self) -> StringOrNone:
        return self.__get("type")

    @property
    def format(self) -> StringOrNone:
        return self.__get("format")

    @property
    def identifier(self) -> StringOrNone:
        return self.__get("identifier")

    @property
    def source(self) -> StringOrNone:
        return self.__get("source")

    @property
    def lang(self) -> StringOrNone:
        return self.__get("lang")

    @property
    def related(self) -> StringOrNone:
        return self.__get("related")

    @property
    def cover(self) -> StringOrNone:
        return self.__get("cover")

    @property
    def rights(self) -> StringOrNone:
        return self.__get("rights")


class Pages:
    """
    Interface for document's pages.

    :Example:

    .. code:: python

       from pyscribus.file import ScribusFile
       from pyscribus.model.pages import Page

       doc = ScribusFile("1.5.5", "example.sla")

       doc_pages = doc.pages

       # Equivalent to doc.pages[0]
       page_one = doc_pages.pages[0]
       how_many_pages = len(doc_pages)
       first_page_copy = doc_pages.copy(0)

       # Master pages are in a attribute of Pages
       master_pages = doc_pages.master_pages
       how_many_master_pages = len(doc_pages.master_pages)

       # BELOW THIS LINE, EVERYTHING IS NOT IMPLEMENTED --------

       doc_pages.remove(5)
       new_page = Page()
       doc_pages.insert(2, new_page)

    """

    def __init__(self, model: SLA):
        self.model = model

    def __len__(self) -> int:
        # > len(x.pages)
        return len(self.__pages_number_sorted("normal"))

    def __getitem__(self, item: int):
        # > x.pages[0]
        # return self.pages[item]
        return list(self.__pages_number_sorted("normal").values())[item]

    @pyfe.has_document
    def __pages_number_sorted(self, page_set="normal") -> OrderedDict:
        if page_set == "normal":
            p_set = self.model.document.pages
        elif page_set == "master":
            p_set = self.model.document.master_pages
        else:
            raise ValueError(f"Asked for invalid page set {page_set}")

        ordered_pages = OrderedDict()
        numbers = sorted([page.number for page in p_set])

        for number in numbers:
            for page in p_set:
                if page.number != number:
                    continue

                ordered_pages[number] = page
                break

        return ordered_pages

    @pyfe.has_document
    def insert(self, position: Union[int, FirstLast], page: Any):
        """
        .. warning:: Not implemented.

        Insert page at page_number ``position``.

        :type position: Union[int, FirstLast]
        :param position: Page number where to insert the page.
        """
        # > x.pages.insert(2, new_page)

        len_pages = len(self)

        insert_last = position == "last" or position == len_pages
        insert_first = position == "first" or position == 0

        if insert_first:
            pass
        if insert_last:
            pass
        else:
            pass

        raise NotImplementedError()

    def copy(
        self,
        position: int,
        new_page_number: Optional[int] = None,
        master_page: bool = False,
    ) -> Union[Page, MasterPage]:
        """
        Returns a copy of the (master)page at ``position`` page_number.

        .. note:: It does not copy page objects, nor does it add the returned
            page to the document.

        :type position: int
        :param position: Page number of the page to copy.
        :type new_page_number: Optional[int]
        :param new_page_number: New page number to attribute to the copy.
        :type master_page: bool
        :param master_page: Copy a master page instead of a regular page.
        :rtype: Union[Page, MasterPage]
        :returns: Copy of the (master)page.
        """
        page_set = {False: "normal", True: "master"}[bool(master_page)]

        try:
            original = list(self.__pages_number_sorted(page_set).values())[
                position
            ]
        except IndexError as not_found:
            raise pyfe.PageNotFound(
                f"There is no page number {position}"
            ) from not_found

        page_copy = copy.deepcopy(original)

        if new_page_number is not None:
            page_copy.number = new_page_number

        return page_copy

    @pyfe.has_document
    def remove(self, page_number: int):
        """
        .. warning:: Not implemented.

        """
        # > x.pages.remove(5)
        # How to :
        # - Catch the coords of the page to delete
        # - Catch the coords of the pages following page_number index
        # - Attribute coords of page_number + 1 the position of page_number + 0
        # - Modify page_number attribute of page_number + 1
        # -   Etc.
        raise NotImplementedError()

    @property
    def pages(self) -> list[Page]:
        # > x.pages
        # return self.model.document.pages
        return list(self.__pages_number_sorted("normal").values())

    @property
    def masterpages(self) -> list[MasterPage]:
        # return self.model.document.master_pages
        return list(self.__pages_number_sorted("master").values())


class Fonts:
    """
    .. warning:: Not implemented.

    Interface for document's fonts.
    """

    # This should inspect styles, Story elements to produce a list/dict of
    # used fonts and their use context.

    def __init__(self, model):
        self.model = model
        self.document = model.document

    def __getitem__(self, item):
        return self.__fonts()[item]

    @pyfe.has_document
    def __fonts(self) -> dict:
        pass


class Layers:
    """
    .. warning:: Not implemented.

    Interface for document's layers.
    """

    def __init__(self, model):
        self.model = model

    def __getitem__(self, item):
        return self.__layers()[item]

    def __iter__(self):
        """
        Special method to make Layers class iterable.
        """

        for layer in self.model.document.layers:
            yield layer

    @pyfe.has_document
    def __layers(self) -> dict:
        return self.model.document.layers

    def insert(self, layer: str, after: int, before: int):
        """
        .. warning:: Not implemented.

        Move layer.
        """

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for colors and gradients.
"""

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

from typing import Union, NoReturn, Literal, List

import lxml
import lxml.etree as ET

import pyscribus.model.common.xml as xmlc
import pyscribus.model.common.math as pmath
import pyscribus.model.exceptions as exceptions
import pyscribus.model.dimensions as dimensions

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

ColorSpace = Literal["cmyk", "rgb"]

# Classes ===============================================================#


class Color(xmlc.PyScribusElement):
    """
    SLA Color (COLOR)

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +------------+---------------------------------+-----------+
    | Kwargs     | Setting                         | Type      |
    +============+=================================+===========+
    | default    | Equivalent to a ``fromdefault`` | boolean   |
    |            | call.                           | or string |
    |            |                                 |           |
    |            | Value being True or the name of |           |
    |            | the default set.                |           |
    +------------+---------------------------------+-----------+
    | name       | Color name                      | string    |
    +------------+---------------------------------+-----------+
    | space      | Color space.                    | string    |
    |            |                                 |           |
    |            | Either "rgb" or "cmyk".         |           |
    +------------+---------------------------------+-----------+
    | cmyk       | Inks in C, M, Y, K order.       | list of   |
    |            |                                 | floats    |
    +------------+---------------------------------+-----------+
    | rgb        | Inks in R, G, B order.          | list of   |
    |            |                                 | floats    |
    +------------+---------------------------------+-----------+
    """

    defaults = {
        "Black": {"space": "cmyk", "colors": [0, 0, 0, 100]},
        "Blue": {"space": "rgb", "colors": [0, 0, 255]},
        "Cool Black": {"space": "cmyk", "colors": [60, 0, 0, 100]},
        "Cyan": {"space": "cmyk", "colors": [100, 0, 0, 0]},
        "Green": {"space": "rgb", "colors": [0, 255, 0]},
        "Magenta": {"space": "cmyk", "colors": [0, 100, 0, 0]},
        "Red": {"space": "rgb", "colors": [255, 0, 0]},
        "Registration": {
            "space": "cmyk",
            "colors": [100, 100, 100, 100],
            "register": True,
        },
        "Rich Black": {"space": "cmyk", "colors": [60, 40, 40, 100]},
        "Warm Black": {
            "space": "cmyk",
            "colors": [0, 60, 29.8039215686275, 100],
        },
        "White": {"space": "cmyk", "colors": [0, 0, 0, 0]},
        "Yellow": {"space": "cmyk", "colors": [0, 0, 100, 0]},
    }

    rgb = lambda s: s.lower() == "rgb"
    cmyk = lambda s: s.lower() == "cmyk"

    def __init__(
        self,
        name: str = "Black",
        space: str = "CMYK",
        colors: list = [0, 0, 0, 100],
        register: str = "0",
        **kwargs,
    ):
        super().__init__()

        self.pyscribus_defaults = [k for k in Color.defaults.keys()]

        self.name = name
        self.register = register

        self.colors = {}
        self.is_cmyk = False
        self.is_rgb = False

        self.set_space_colors(space, colors)

        self._quick_setup(kwargs)

    # Color management -----------------------------------------

    @property
    def space(self):
        try:
            return {True: "cmyk", False: "rgb"}[self.is_cmyk]
        except KeyError:
            return None

    def set_space_colors(
        self, space: ColorSpace, colors: List[float]
    ) -> NoReturn:
        """
        Set color space and color inks of the colors.

        :type space: string
        :param space: Color space. Either "cmyk" or "rgb".
        :type colors: list
        :param colors: List of inks values (float).

        :Example:

        .. code:: python

           # Green RGB
           color.set_space_colors("rgb", [0, 255, 0])

           # White CMYK
           color.set_space_colors("cmyk", [0, 0, 0, 0])

        """

        spaced = self.set_space(space)

        if spaced:
            self.set_colors(colors, self.space)

    def set_colors(
        self, colors: List[float], space: Union[ColorSpace, bool]
    ) -> bool:
        """
        :type colors: list
        :param colors: List of inks values (float).
        :type space: string
        :param space: Optionnal. Color space. Either "cmyk" or "rgb".
        :rtype: bool
        """

        if not space:
            space = {True: "cmyk", False: "rgb"}[self.is_cmyk]

        if space.lower() in ["cmyk", "rgb"]:

            if Color.cmyk(space):

                self.colors = {
                    "C": float(colors[0]),
                    "M": float(colors[1]),
                    "Y": float(colors[2]),
                    "K": float(colors[3]),
                }

                # Avoid invalid values

                for ink in ["C", "M", "Y", "K"]:

                    if self.colors[ink] > float(100.0):
                        self.colors[ink] = 100.0

                    if self.colors[ink] < float(0.0):
                        self.colors[ink] = 0.0

            if Color.rgb(space):
                self.colors = {
                    "R": float(colors[0]),
                    "G": float(colors[1]),
                    "B": float(colors[2]),
                }

                # Avoid invalid values

                for ink in ["R", "G", "B"]:
                    if self.colors[ink] > float(255.0):
                        self.colors[ink] = 255.0

                for ink in ["R", "G", "B"]:
                    if self.colors[ink] < float(0.0):
                        self.colors[ink] = 0.0

            return True

        return False

    def set_space(self, space: ColorSpace) -> bool:
        """
        Set the color space (CMYK / RGB) of the color.

        :type space: string
        :param space: Color space. Either "cmyk" or "rgb".
        :rtype: bool
        """

        space = space.lower()

        if space in ["cmyk", "rgb"]:
            self.is_cmyk = Color.cmyk(space)
            self.is_rgb = Color.rgb(space)

            return True

        return False

    # PyScribus standard methods -------------------------------

    def fromdefault(self, default: str = "Black") -> bool:
        """
        Set default attributes for this color.

        :Example:

        .. code:: python

           # with fromdefault method :
           black = Color()
           black.fromdefault("Black")

           # with quick setup :
           blue = Color(default="Blue")


        .. seealso::
            **For fromdefault explanation**:
            `pyscribus.common.xml.PyScribusElement.fromdefault`

            **For current class default sets**:
            `pyscribus.colors.Color.listdefaults`

        """

        if default in self.pyscribus_defaults:

            self.name = default

            self.set_space_colors(
                Color.defaults[default]["space"],
                Color.defaults[default]["colors"],
            )

            if "register" in Color.defaults[default]:
                self.register = True

            return True

        return False

    def fromxml(self, xml: ET.Element) -> bool:
        """
        :type xml: lxml.etree._Element
        :param xml: XML source of color
        :rtype: boolean
        :returns: True if XML parsing succeed
        """

        space = xml.get("SPACE")

        if space is None:
            return False

        # Name -------------------------------------------------------

        if (name := xml.get("NAME")) is not None:
            self.name = name

        # Space ------------------------------------------------------

        if space.lower() in ["cmyk", "rgb"]:
            self.set_space(space)

            # Colors -------------------------------------------------

            if self.is_cmyk:
                colors = [xml.get(c) for c in ["C", "M", "Y", "K"]]
            else:
                colors = [xml.get(c) for c in ["R", "G", "B"]]

            if not colors:
                raise exceptions.InvalidColor("No inks.")

            if None in colors:
                raise exceptions.InvalidColor(
                    "Invalid inks <{}>.".format(",".join(colors))
                )

            self.set_colors(colors, self.space)

        # Register ---------------------------------------------------

        if (reg := xml.get("Register")) is None:
            self.register = False
        else:
            if int(reg):
                self.register = True
            else:
                self.register = False

        # ------------------------------------------------------------

        return True

    def toxml(self) -> ET.Element:
        """
        :rtype: lxml.etree._Element
        :returns: Color as XML element
        """

        xml = ET.Element("COLOR")
        xml.attrib["NAME"] = self.name

        if self.is_cmyk:
            xml.attrib["SPACE"] = "CMYK"

            for color in ["C", "M", "Y", "K"]:
                color_value = pmath.necessary_float(self.colors[color])
                xml.attrib[color] = str(color_value)

        else:
            xml.attrib["SPACE"] = "RGB"

            for color in ["R", "G", "B"]:
                color_value = pmath.necessary_float(self.colors[color])
                xml.attrib[color] = str(color_value)

        if self.register:
            xml.attrib["Register"] = "1"

        return xml

    def _quick_setup(self, settings: dict) -> NoReturn:
        """
        Method for defining gradient stop settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            xmlc.PyScribusElement._quick_setup(self, settings)

            # Color space is treated first. ------------------------------

            if "space" in settings:
                self.set_space(str(settings["space"]))
                settings.pop("space")

            for setting_name, setting_value in settings.items():

                # Color name ---------------------------------------------

                if setting_name == "name":
                    self.name = str(setting_value)

                # Setting colors -----------------------------------------

                if setting_name == "rgb":
                    if len(setting_value) < 3:
                        while len(setting_value) < 3:
                            setting_value.append(float(0))

                    setting_value = setting_value[:3]
                    self.set_colors(setting_value, self.space)

                if setting_name == "cmyk":
                    if len(setting_value) < 4:
                        while len(setting_value) < 4:
                            setting_value.append(float(0))

                    setting_value = setting_value[:4]
                    self.set_colors(setting_value, self.space)

    def same_space(self, other: Color) -> bool:
        """
        Is that color and other color share the same color space ?

        :param other: Color instance to compare.
        :type other: pyscribus.colors.Color
        :returns: True if both color share the same color space.
        :rtype: bool
        """

        # NOTE Obviously we don't translate colors of different spaces
        # to compare their inks, as RVB and CMYK color spectrums are
        # not the same.

        if self.is_cmyk and other.is_cmyk:
            return True
        if self.is_rgb and other.is_rgb:
            return True

        return False

    # Python __ methods ----------------------------------------

    def __iadd__(self, other: Color) -> Color:
        if not self.same_space(other):
            raise exceptions.ColorMixing(
                f"Colors {self} and {other} do not share the same color space."
            )

        # FIXME TODO Implement that !

        return self

    def __eq__(self, other: Color) -> bool:
        """
        Equality operator.

        Doesn’t check if RVB color can be an equivalent in CMYK.
        """

        if not self.same_space(other):
            return False

        # NOTE Self colors and other colors are in the same space, so
        # inks dicts have the same keys. So we get a list of
        # [(R1,R2)…] or [(C1,C2)…]. Having one ink different is enough.
        inks = zip(self.colors.values(), other.values())

        for ink in inks:
            if ink[0] != ink[1]:
                return False

        return True

    def __repr__(self):
        space_name = {True: "CMYK", False: "RGB"}[self.is_cmyk]
        inks = {True: ["C", "M", "Y", "K"], False: ["R", "G", "B"]}[
            self.is_cmyk
        ]

        return "{}:{}:{}:{}".format(
            self.name,
            space_name,
            ";".join([str(self.colors[ink]) for ink in inks]),
            self.register,
        )


class GradientColorStop(xmlc.PyScribusElement):
    """
    Gradient color stop (Gradient/CSTOP)

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +------------+---------------------------------+-----------+
    | Kwargs     | Setting                         | Type      |
    +============+=================================+===========+
    | default    | Equivalent to a ``fromdefault`` | boolean   |
    |            | call.                           | or string |
    |            |                                 |           |
    |            | Value being True or the name of |           |
    |            | the default set.                |           |
    +------------+---------------------------------+-----------+
    | opacity    | Percentage of opacity,          | float     |
    |            | from 0 to 1.                    |           |
    +------------+---------------------------------+-----------+
    | name       | Color name                      | string    |
    +------------+---------------------------------+-----------+
    | position   |                                 |           |
    +------------+---------------------------------+-----------+
    | shade      |                                 |           |
    +------------+---------------------------------+-----------+

    :Example:

    .. code:: python

       cstop1 = colors.GradientColorStop(
           color="Black", shade=100, position=0, opacity=1
       )

    """

    def __init__(self, **kwargs):
        super().__init__()

        self.name = None
        self.shade = None
        self.opacity = None
        self.position = None

        self._quick_setup(kwargs)

    def __eq__(self, other) -> bool:
        """
        Equality operator.
        """

        def compare_property(a, b):
            if a is None:
                return b is not None

            if other.position is not None:
                return a.value != b.value

            return False

        if self.name != other.name:
            return False

        if compare_property(self.position, other.position):
            return False

        if compare_property(self.shade, other.shade):
            return False

        if compare_property(self.opacity, other.opacity):
            return False

        return True

    def _quick_setup(self, settings: dict) -> NoReturn:
        """
        Method for defining gradient stop settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            xmlc.PyScribusElement._quick_setup(self, settings)

            for setting_name, setting_value in settings.items():

                if setting_name == "name":
                    self.name = setting_value

                if setting_name == "position":
                    self.position = dimensions.Dim(
                        float(setting_value), unit="pcdecim"
                    )

                if setting_name == "opacity":
                    self.opacity = dimensions.Dim(
                        float(setting_value), unit="pcdecim"
                    )

                if setting_name == "shade":
                    self.shade = dimensions.Dim(float(setting_value), unit="pc")

    def fromdefault(self) -> NoReturn:
        self.opacity = dimensions.Dim(100, unit="pc")
        self.position = dimensions.Dim(0, unit="pcdecim")

    def fromxml(self, xml: ET._Element):
        if xml.tag == "CSTOP":
            if (name := xml.get("NAME")) is not None:
                self.name = name

            if (shade := xml.get("SHADE")) is not None:
                self.shade = dimensions.Dim(float(shade), unit="pc")

            if (ramp := xml.get("RAMP")) is not None:
                self.position = dimensions.Dim(float(ramp), unit="pcdecim")

            if (opacity := xml.get("TRANS")) is not None:
                self.opacity = dimensions.Dim(float(opacity), unit="pcdecim")

            return True

        return False

    def toxml(self) -> ET.Element:
        xml = ET.Element("CSTOP")

        if self.position is not None:
            xml.attrib["RAMP"] = self.position.toxmlstr()

        xml.attrib["NAME"] = self.name

        if self.shade is None:
            xml.attrib["SHADE"] = "100"
        else:
            xml.attrib["SHADE"] = self.shade.toxmlstr(True)

        if self.opacity is None:
            xml.attrib["TRANS"] = "1"
        else:
            xml.attrib["TRANS"] = self.opacity.toxmlstr(True)

        return xml


class Gradient(xmlc.PyScribusElement):
    """
    Gradient in SLA (Gradient)
    """

    # <Gradient Name="Orange, Jaune" Ext="3">
    # </Gradient>

    def __init__(self):
        super().__init__()

        self.name = None
        self.stops = []

    def _sorted_stops(self) -> list:
        return sorted(self.stops, key=lambda s: s.position.value)

    def fromxml(self, xml: ET.Element) -> bool:
        if xml.tag == "Gradient":

            if (name := xml.get("Name")) is not None:
                self.name = name

            # TODO FIXME Ext

            for element in xml:
                if element.tag == "CSTOP":
                    grs = GradientColorStop()

                    if success := grs.fromxml(element):
                        self.stops.append(grs)

            return True

        return False

    def append_stop(self, stop: GradientColorStop, sort: bool = False) -> bool:
        """
        Append a gradient color stop.

        Avoids duplicates. Can sort stops by color stop position.

        :type stop: pyscribus.colors.GradientColorStop
        :param stop: Gradient color stop to append
        :type sort: boolean
        :param sort: Sort gradient color stops by position.
        """

        if isinstance(stop, GradientColorStop):

            duplicate = [cs for cs in self.stops if cs == stop]

            if not duplicate:

                self.stops.append(stop)

                if sort:
                    self.stops = self._sorted_stops()

                return True

        return False

    def toxml(self) -> ET.Element:
        xml = ET.Element("Gradient")

        xml.attrib["Name"] = self.name

        # TODO FIXME Ext

        if self.stops:

            # Sort the color stops of the gradient by their position in
            # the gradient spectrum, ranging from 0 to 100%

            for stop in self._sorted_stops():
                stop_xml = stop.toxml()
                xml.append(stop_xml)

        return xml


# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for notes
"""

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

import copy

from typing import Union, NoReturn

import lxml
import lxml.etree as ET

import pyscribus.model.exceptions as exceptions

from pyscribus.model.common.xml import *

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#


class Note(PyScribusElement):
    """
    Note element (Note)
    """

    def __init__(self, doc_parent=False):
        PyScribusElement.__init__(self)

        self.style = ""
        self.parent_mark = ""
        self.parent_frame = ""

        self.doc_parent = doc_parent

        self.text = NoteText(parent_note=self)

    def fromdefault(self) -> NoReturn:
        self.style = "Default"
        self.text = NoteText()
        self.text.fromdefault()

    def get_parent_mark(self) -> str:
        return "NoteMark_{} in frame {}".format(self.style, self.parent_frame)

    def fromxml(self, xml: ET.Element) -> bool:

        if xml.tag != "Note":
            return False

        style = xml.get("NStyle")
        parent_mark = xml.get("Master")

        if style is not None:
            self.style = style

        if parent_mark is not None:
            self.parent_mark = parent_mark
            self.parent_frame = self.parent_mark.split("in frame ")[-1]

        if "Text" in xml.attrib:
            text = NoteText()
            success = text.fromxml(xml)

            if success:
                self.text = text

        return True

    def toxml(self) -> ET.Element:
        xml = ET.Element("Note")

        xml.attrib["Master"] = self.get_parent_mark()
        xml.attrib["NStyle"] = self.style
        xml.attrib["Text"] = self.text.toxmlstr()

        return xml


class NoteText(PyScribusElement):
    """
    Note element text (Note/@Text).

    Note element text consists in a full XML file dumped into @Text attribute
    of Note, hence the need for a specific class.
    """

    # Text="&lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;&lt;SCRIBUSTEXT &gt;&lt;defaultstyle /&gt;&lt;p &gt;&lt;style /&gt;&lt;span &gt;&lt;charstyle Features=&quot;inherit &quot; /&gt; Pri tiuj ĉi ideoj « estas modo » paroli ne alie, ol kun ironia kaj malestima rideto, tial tiel agas ankaŭ A kaj B kaj C, kaj ĉiu el ili timas enpensiĝi serioze eĉ unu minuton pri la mokata ideo, ĉar li « scias antaŭe », ke « ĝi krom malsaĝaĵo enhavas ja nenion », kaj li timas, ke oni iel alkalkulos lin mem al la nombro de « tiuj malsaĝuloj », se li eĉ en la daŭro de unu minuto provos rilati serioze al tiu ĉi malsaĝaĵo. La homoj miras, « kiamaniere en nia praktika tempo povas aperi tiaj malsaĝaj fantaziuloj kaj kial oni ne metas ilin en la domojn por frenezuloj ».&lt;/span&gt;&lt;/p&gt;&lt;/SCRIBUSTEXT&gt;&#10;"/>

    # Text="
    # <?xml version="1.0" encoding="UTF-8"?>
    # <SCRIBUSTEXT >
    # <defaultstyle />
    # <p >
    # <style />
    # <span >
    # <charstyle Features="inherit " />
    # Pri tiuj ĉi ideoj « estas modo » paroli ne alie, ol kun ironia kaj malestima rideto, tial tiel agas ankaŭ A kaj B kaj C, kaj ĉiu el ili timas enpensiĝi serioze eĉ unu minuton pri la mokata ideo, ĉar li « scias antaŭe », ke « ĝi krom malsaĝaĵo enhavas ja nenion », kaj li timas, ke oni iel alkalkulos lin mem al la nombro de « tiuj malsaĝuloj », se li eĉ en la daŭro de unu minuto provos rilati serioze al tiu ĉi malsaĝaĵo. La homoj miras, « kiamaniere en nia praktika tempo povas aperi tiaj malsaĝaj fantaziuloj kaj kial oni ne metas ilin en la domojn por frenezuloj ».
    # </span>
    # </p>
    # </SCRIBUSTEXT>
    # &#10;"/>

    empty_text = '&lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;' \
    '&lt;SCRIBUSTEXT &gt;&lt;defaultstyle /&gt;' \
    '&lt;p &gt;&lt;style /&gt;&lt;span &gt;' \
    '&lt;charstyle Features=&quot;inherit &quot; /&gt;' \
    '&lt;/span&gt;&lt;/p&gt;&lt;/SCRIBUSTEXT&gt;&#10;'

    empty_pattern = '&lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;' \
    '&lt;SCRIBUSTEXT &gt;&lt;defaultstyle /&gt;' \
    '&lt;p &gt;&lt;style /&gt;&lt;span &gt;' \
    '&lt;charstyle Features=&quot;inherit &quot; /&gt;' \
    '&lt;/span&gt;&lt;/p&gt;{0}&lt;/SCRIBUSTEXT&gt;&#10;'

    def __init__(self, parent_note=False):
        PyScribusElement.__init__(self)

        self.xml = None

        # We need to link NoteText to its note element (Note), because Note
        # is linked with pyscribus.document.Document.
        self.parent_note = parent_note

    def fromxml(self, xml: ET.Element) -> bool:

        if xml.tag != "Note":
            return False

        content = xml.get("Text")

        if content is None:
            return False

        text = content

        # &#10; = Line feed character

        for normalise in [
            ["&gt;&lt;", "><"],
            [" &gt;", ">"],
            ["&lt;", "<"],
            ["&quot;", '"'],
            ["&#10;", ""],
        ]:
            text = text.replace(normalise[0], normalise[1])

        # Remove encoding declaration because lxml.etree.fromstring does not
        # support it in unicode strings.
        text = text.replace('<?xml version="1.0" encoding="UTF-8"?>', "")

        print(text)

        self.xml = ET.fromstring(text)

        return True

    def toxml(self) -> str:
        """
        Alias of toxmlstr()
        """
        return self.toxmlstr()

    def toxmlstr(self) -> str:
        if self.xml is None:
            return NoteText.empty_text

        xml_string = copy.deepcopy(self.xml)
        xml_string = ET.tostring(xml_string, encoding="utf8")

        xml_string = xml_string.replace("><", "&gt;&lt;")
        xml_string = xml_string.replace(">", "&gt;")
        xml_string = xml_string.replace("<", "&lt;")
        xml_string = xml_string.replace('"', "&quot;")
        xml_string += "&#10;"

        print(xml_string)

        return xml_string


class NoteFrame(PyScribusElement):
    """
    Note frame (NotesFrames/FOOTNOTEFRAME).
    """

    def __init__(self):
        PyScribusElement.__init__(self)

        # NOTE @myID
        # The page object where the note content is located
        self.own_frame_id = None

        # NOTE @MasterID
        # The page object with the story where note references are located
        self.story_frame_id = None

        # NOTE @NSname
        self.note_style = None

    def fromxml(self, xml: ET.Element) -> bool:

        if xml.tag != "FOOTNOTEFRAME":
            return False

        note_style = xml.get("NSname")
        own_frame = xml.get("myID")
        story_frame = xml.get("MasterID")

        if note_style is not None:
            self.note_style = note_style

        if own_frame is not None:
            self.own_frame_id = own_frame

        if story_frame is not None:
            self.story_frame_id = story_frame

        if story_frame is None or own_frame is None:

            raise exceptions.InsaneSLAValue(
                "Note frame has no page object ID or\
                 no parent page object ID."
            )

        return True

    def toxml(self) -> ET.Element:
        xml = ET.Element("FOOTNOTEFRAME")

        if self.note_style is not None:
            xml.attrib["NSname"] = self.note_style

        if self.own_frame_id is not None:
            xml.attrib["myID"] = self.own_frame_id

        if self.story_frame_id is not None:
            xml.attrib["MasterID"] = self.story_frame_id

        return xml

    def fromdefault(self) -> NoReturn:
        self.note_style = "Default"


# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus

Scribus file. Colors and gradients wrappers.
"""

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

from pyscribus.model.sla import SLA
from pyscribus.model.colors import Color
import pyscribus.file.exceptions as pyfe

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#


class Colors:
    """
    Wrapper around Scribus colors.

    :type file: pyscribus.file.ScribusFile
    :param file: Scribus file wrapper

    :Example:

    .. code:: python

       from pyscribus.file import ScribusFile

       doc = ScribusFile("1.5.5", "example.sla")

       how_many_colors = len(doc.colors)
       black_color = doc.colors["Black"]
       cmyk_colors = doc.colors.filter(space="cmyk")

    """

    def __init__(self, model: SLA):
        self.model = model

    def __len__(self) -> int:
        # > len(x.colors)
        return len(self.__colors())

    def __getitem__(self, item: int):
        # > x.colors["black"]
        return self.__colors()[item]

    @pyfe.has_document
    def __colors(self) -> dict:
        return {color.name: color for color in self.model.document.colors}

    @pyfe.has_document
    def filter(self, **kwargs) -> list[Color]:
        """
        Returns a selection of colors.

        :type kwargs: dict
        :param kwargs: kwargs
        :rtype: list[Color]
        :returns: Selection of colors

        **Kwargs options :**

        +-----------------------+-------------+-------------------------+
        | Kwarg key             | Filter on   | Type                    |
        +=======================+=============+=========================+
        | space                 | Color space | ``"cmyk"`` or ``"rgb"`` |
        +-----------------------+-------------+-------------------------+
        """

        colors = self.__colors()

        filter_space = False

        if (space := kwargs.get("space")) is not None:
            if space.lower() in ["cmyk", "rgb"]:
                filter_space = space.lower()

        if filter_space:
            colors = {
                color.name: color
                for color in colors.values()
                if color.space == filter_space
            }

        return colors

    @property
    def items(self) -> dict:
        """
        Colors dictionary organized by color name.

        .. seealso:: :class:`pyscribus.model.colors.Color`

        :rtype: dict
        """
        return self.__colors()


# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus module for PDF & printing settings / elements.
"""

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

import collections

from typing import NoReturn, Union

import lxml
import lxml.etree as ET

import pyscribus.model.exceptions as exceptions

from pyscribus.model.common.xml import *

import pyscribus.model.dimensions as dimensions

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#


class PDFSettings(PyScribusElement):
    """
    Class for PDF export settings in SLA
    """

    imgcomp_method_xml = {"automatic": 0, "jpeg": 1, "flate": 2, "none": 3}

    imgcomp_quality_xml = {
        "max": 0,
        "high": 1,
        "medium": 2,
        "low": 3,
        "min": 4,
    }

    def __init__(self):
        super().__init__()

        self.lpi = []
        self.use_lpi = False

        self.text_compression = True
        self.image_compression = {"method": "automatic", "quality": "max"}

        # ------------------------------------------------------

        # Flag, set to 1 when Presentation effects should be used
        PresentMode = "0"

        # ------------------------------------------------------

        # Flag, set to 1 when Downsample Images is checked in the PDF-Options Dialog
        RecalcPic = "0"
        # Resolution for downsampling Images
        PicRes = "300"
        # Resolution for embedded EPS-Pictures or PDF's
        Resolution = "300"

        # ------------------------------------------------------

        # (optional) Flag, set to 1 when Output should be in RGB
        RGBMode = "1"

        # TODO : fromxml / toxml of this dict

        self.profiles = {
            "colors": {
                # (optional) Flag, set to 1 when ICC-Profiles should be used
                # for solid colours
                # UseProfiles="0"
                "used": False,
                # (optional) ICC-Profile for solid colours
                # SolidP="sRGB IEC61966-2.1"
                "name": "sRGB IEC61966-2.1",
            },
            "images": {
                # (optional) Flag, set to 1 when ICC-Profiles should be used
                # for images
                # UseProfiles2="0"
                "used": False,
                # ICC-Profile for images ?
                # ImageP="sRGB IEC61966-2.1"
                "name": "sRGB IEC61966-2.1",
            },
            "printer": {
                # (optional) ICC-Profile for the Printer, only used for
                # PDF-X/3 output
                # PrintP="Fogra27L CMYK Coated Press"
                "name": "Fogra27L CMYK Coated Press"
            },
        }

        # ------------------------------------------------------

        self.bleeds = collections.OrderedDict()
        self.bleeds["top"] = dimensions.Dim(0)
        self.bleeds["left"] = dimensions.Dim(0)
        self.bleeds["right"] = dimensions.Dim(0)
        self.bleeds["bottom"] = dimensions.Dim(0)
        self.bleeds["document"] = False

        # ------------------------------------------------------

        self.marks = {
            # bool as 0 or 1

            # bleedMarks (dimensions of bleeds in self.bleeds)
            "bleeds": False,
            # cropMarks, cuting marks
            "croping": False,
            # docInfoMarks, page infos
            "infos": False,
            # colorMarks, color reference
            "colors": False,
            # registrationMarks, page alignment
            "alignment": False,

            # pica dims

            # markOffset
            "offset": dimensions.Dim(0),
            # markLength, 7.056mm but encoded in pics => roughly 20
            "length": dimensions.Dim(20),
        }

        # ------------------------------------------------------

        # All         : -2359300
        # None        : -2359360
        # Only print  : -2359356
        # Only modify : -2359352
        # Only copy   : -2359344
        # Only annoto : -2359328

        self.encryption = {
            "pass": {"owner": "", "user": ""},
            "settings": {"permissions": "-4", "encrypted": False},
        }

        # FIXME Documented but to organize ----------------------

        # Binding for the PDF-Document 0 = Left Binding 1 = Right Binding
        Binding = "0"

        # PDF version -------------------------------------------

        self.version = "13"

        # FIXME Documented but to organize ----------------------

        self.objects = {
            # Flag, set to 1 when Generate Thumbnails is checked in the
            # PDF-Options Dialog
            # Thumbnails = "0"
            "thumbnails": False,
            # Flag, set to 1 when include Bookmarks is checked in the
            # PDF-Options Dialog
            # Bookmarks = "0"
            "bookmarks": False,
            # Flag, set to 1 when use PDF-Articles is checked in the
            # PDF-Options Dialog
            # Articles = "0"
            "articles": False,
            # EmbedPDF="0"
            "embedded_pdf": False,
        }

        # FIXME Undocumented, managed with undocumented funs ----

        # ImagePr="0"
        # UseLayers="0"
        # UseSpotColors="1"
        # doMultiFile="0"
        # displayBookmarks="0"
        # displayFullscreen="0"
        # displayLayers="0"
        # displayThumbs="0"
        # hideMenuBar="0"
        # hideToolBar="0"
        # fitWindow="0"
        # openAfterExport="0"
        # PageLayout="0"
        # openAction=""
        # Intent="1"
        # Intent2="0"
        # InfoString=""
        # FontEmbedding="0"
        # Grayscale="0"
        # firstUse="1"
        # Clip="0"
        # rangeSel="0"
        # rangeTxt=""

        # Transformation ----------------------------------------

        self.transformation = {
            # RotateDeg="0"
            # Four values: 0, 90, 180, 270
            "rotation": dimensions.Dim(0, unit="deg", integer=True),
            # MirrorH="0"
            "mirror_h": False,
            # MirrorV="0"
            "mirror_v": False,
        }

    def set_version(self, version: Union[str, int, float]):
        # We should check against Scribus version because
        # PDF-X-3, PDF-X-4, PDF-X-1a seems not available in Scribus 1.5.

        # 12 = PDF-X/3
        # 13 = PDF-1.3 (Acrobat 4)
        # 14 = PDF-1.4 (Acrobat 5)
        # 15 = PDF-1.5 (Acrobat 6)
        # 16 = PDF-1.6 (Acrobat 7)

        new_version = None

        if isinstance(version, float):

            if int(version) in [13, 14, 15, 16]:
                self.version = str(int(version))
                return True

            if version in [1.3, 1.4, 1.5, 1.6]:
                new_version = str(version)
                self.version = new_version.replace(".", "")
                return True

        if isinstance(version, int):

            if version in [13, 14, 15, 16]:
                self.version = str(version)
                return True

        if isinstance(version, str):

            if version in ["1.3", "1.4", "1.5", "1.6"]:
                new_version = version
                self.version = new_version.replace(".", "")
                return True

            if version.startswith("acrobat-"):
                acro_version = version.split("acrobat-")[-1]

                if not acro_version:
                    return False

                try:
                    self.version = {
                        "4": "13",
                        "5": "14",
                        "6": "15",
                        "7": "16",
                    }[acro_version]
                    return True
                except KeyError:
                    return False

            if version == "X/3":
                self.version = "12"
                return True

        return False

    def fromxml(self, xml: ET.Element) -> bool:

        if xml.tag != "PDF":
            return False

        # TODO

        text_comp = xml.get("Compress")

        if text_comp is not None:
            self.text_compression = num_to_bool(text_comp)

        # Image compression : method and quality -----------

        for sett in [
            [xml.get("CMethod"), "method", PDFSettings.imgcomp_method_xml],
            [xml.get("Quality"), "quality", PDFSettings.imgcomp_quality_xml],
        ]:

            if (att := xml.get(sett[0])) is not None:
                if int(att) in sett[2].values():
                    for human, code in sett[2].items():
                        self.image_compression[sett[1]] = human
                        break
                else:
                    raise exceptions.InsaneSLAValue(
                        "Unknown image compression {}.".format(sett[1])
                    )

        # Bleed settings -----------------------------------

        for edge in ["top", "left", "right", "bottom"]:
            att_name = f"B{edge.capitalize()}"

            if (att := xml.get(att_name)) is not None:
                self.bleeds[edge].value = float(att)

        udb = xml.get("useDocBleeds")

        if udb is not None:
            self.bleeds["document"] = num_to_bool(udb)

        # PDF marks settings -------------------------------

        for mark_tag, mark_key in [
                ["bleedMarks", "bleeds"],
                ["cropMarks", "croping"],
                ["docInfoMarks", "infos"],
                ["colorMarks", "colors"],
                ["registrationMarks", "alignment"]]:

            if (attrib := xml.get(mark_tag)) is not None:
                self.marks[mark_key] = num_to_bool(attrib)

        for mark_tag, mark_key in [
                ["markOffset", "offset"],
                ["markLength", "length"]]:
            if (attrib := xml.get(mark_tag)) is not None:
                self.marks[mark_key].value = float(attrib)

        # Version --------------------------------------------------------

        version = xml.get("Version")

        if version is not None:
            self.version = version

        # Objects to embed / generate ------------------------------------

        embedded_pdf = xml.get("EmbedPDF")
        thumbnails = xml.get("Thumbnails")
        bookmarks = xml.get("Bookmarks")
        articles = xml.get("Articles")

        for att_value, human in [
                [thumbnails, "thumbnails"], [bookmarks, "bookmarks"],
                [articles, "articles"], [embedded_pdf, "embedded_pdf"]]:
            if att_value is None:
                continue

            self.objects[human] = num_to_bool(att_value)

        # Transformation ----------------------------------------

        rotation = xml.get("RotateDeg")

        if rotation is not None:
            try:
                rotation = int(rotation)

                if rotation in [0, 90, 180, 270]:
                    self.transformation["rotation"].value = rotation
                else:
                    raise exceptions.InsaneSLAValue(
                        "PDF settings Rotation: "
                        "Allowed values are 0, 90, 180, 270. "
                        f"Got {rotation}."
                    )
            except ValueError:
                raise exceptions.InsaneSLAValue(
                    "PDF settings Rotation: "
                    "Requires an integer. "
                    f"Got {rotation}."
                )

        mirrorh = xml.get("MirrorH")
        mirrorv = xml.get("MirrorV")

        for att_value, human in [
                [mirrorh, "mirror_h"],  [mirrorv, "mirror_v"]]:
            if att_value is None:
                continue

            self.transformation[human] = num_to_bool(att_value)

        # PDF encryption settings-----------------------------------------

        encrypt = xml.get("Encrypt")
        encrypt_perm = xml.get("Permissions")

        for pass_key, pass_tag in [["owner", "PassOwner"], ["user", "PassUser"]]:
            if (attrib := xml.get(pass_tag)) is not None:
                self.encryption["pass"][pass_key] = attrib

        if encrypt is not None:
            self.encryption["settings"]["encrypted"] = num_to_bool(encrypt)

        if encrypt_perm is not None:
            self.encryption["settings"]["permissions"] = encrypt_perm

        # Line per inch settings ---------------------------

        # Flag, set to 1 when the informations in the LPI tags should be used
        # for Linescreening
        # IT IS NOT OPTIONAL
        uselpi = xml.get("UseLpi")
        if uselpi is not None:
            self.use_lpi = num_to_bool(uselpi)

        for element in xml:
            if element.tag != "LPI":
                continue

            lpi_item = LPI()
            success = lpi_item.fromxml(element)

            if success:
                self.lpi.append(lpi_item)

        # FIXME This records undocumented attributes ----------

        self.undocumented = undocumented_to_python(
            xml,
            [
                "ImagePr",
                "UseLayers",
                "UseSpotColors",
                "doMultiFile",
                "displayBookmarks",
                "displayFullscreen",
                "displayLayers",
                "displayThumbs",
                "hideMenuBar",
                "hideToolBar",
                "fitWindow",
                "openAfterExport",
                "PageLayout",
                "openAction",
                "Intent",
                "Intent2",
                "InfoString",
                "FontEmbedding",
                "Grayscale",
                "firstUse",
                "Clip",
                "rangeSel",
                "rangeTxt",
            ],
        )

        return True

    def toxml(self) -> ET.Element:
        xml = ET.Element("PDF")

        xml.attrib["Compress"] = bool_to_num(self.text_compression)

        xml.attrib["CMethod"] = str(
            PDFSettings.imgcomp_method_xml[self.image_compression["method"]]
        )

        xml.attrib["Quality"] = str(
            PDFSettings.imgcomp_quality_xml[self.image_compression["quality"]]
        )

        # Bleed settings -----------------------------------

        for bleed_config, bleed_value in self.bleeds.items():

            if bleed_config != "document":
                att_name = "B{}".format(bleed_config.capitalize())
                # FIXME Sometimes raises errors (when PDF/@BTop="0") because
                # not a list
                try:
                    xml.attrib[att_name] = bleed_value[0].toxmlstr(True)
                except:
                    xml.attrib[att_name] = bleed_value.toxmlstr(True)

        xml.attrib["useDocBleeds"] = bool_to_num(self.bleeds["document"])

        # PDF marks settings -------------------------------

        for mark_tag, mark_key in [
                ["bleedMarks", "bleeds"],
                ["cropMarks", "croping"],
                ["docInfoMarks", "infos"],
                ["colorMarks", "colors"],
                ["registrationMarks", "alignment"]]:
            xml.attrib[mark_tag] = bool_to_num(self.marks[mark_key])

        for mark_tag, mark_key in [
                ["markOffset", "offset"],
                ["markLength", "length"]]:
            xml.attrib[mark_tag] = self.marks[mark_key].toxmlstr(True)

        # PDF encryption settings---------------------------

        xml.attrib["PassOwner"] = self.encryption["pass"]["owner"]
        xml.attrib["PassUser"] = self.encryption["pass"]["user"]
        xml.attrib["Permissions"] = self.encryption["settings"]["permissions"]
        xml.attrib["Encrypt"] = bool_to_num(
            self.encryption["settings"]["encrypted"]
        )

        # Version ------------------------------------------

        xml.attrib["Version"] = str(self.version)

        # Objects to embed / generate ------------------------------------

        for code, human in [
                ["Thumbnails", "thumbnails"], ["Bookmarks", "bookmarks"],
                ["Articles", "articles"], ["EmbedPDF", "embedded_pdf"]]:
            xml.attrib[code] = bool_to_num(self.objects[human])

        # Transformation ----------------------------------------

        xml.attrib["RotateDeg"] = self.transformation["rotation"].toxmlstr(True)

        for code, human in [["MirrorH", "mirror_h"], ["MirrorV", "mirror_v"]]:
            xml.attrib[code] = bool_to_num(self.transformation[human])

        # Line per Inch children ---------------------------

        xml.attrib["UseLpi"] = bool_to_num(self.use_lpi)

        for lpi_item in self.lpi:
            lpi_xml = lpi_item.toxml()

            if lpi_xml is not None:
                xml.append(lpi_xml)

        # FIXME This exports undocumented attributes ----------

        try:
            xml = undocumented_to_xml(xml, self.undocumented)
        except AttributeError:
            pass

        return xml

    def fromdefault(self) -> NoReturn:
        self.text_compression = True
        self.image_compression = {"method": "automatic", "quality": "max"}

        # Default line per each settings

        self.lpi = []

        for dlpi in LPI.DEFAULTS:
            lobj = LPI()
            lobj.fromdefault(dlpi[0])

            self.lpi.append(lobj)


class LPI(PyScribusElement):
    """
    Lines per Inch settings (LPI) in SLA
    """

    spot_xml = {
        "dot": "0",
        "line": "1",
        "round": "2",
        "ellipse": "3",
    }

    DEFAULTS = [
        ["Black", 133, 45, "ellipse"],
        ["Cyan", 133, 105, "ellipse"],
        ["Magenta", 133, 75, "ellipse"],
        ["Yellow", 133, 90, "ellipse"],
    ]

    def __init__(self):
        super().__init__()

        self.pyscribus_defaults = [lpi_def[0] for lpi_def in LPI.DEFAULTS]

        # Linescreening angle
        self.angle = dimensions.Dim(0, unit="deg")

        # Name of the Colour for which these settings are ment
        self.color = ""

        # How many lines per Inch are used
        self.frequency = dimensions.Dim(133, unit="lpi", integer=True)

        # Code for the used Spotfunction
        self.spot = "ellipse"

    # PyScribus standard methods -------------------------------

    def fromdefault(self, default: str) -> NoReturn:

        if default in self.pyscribus_defaults:

            self.color = default

            for lpd in LPI.DEFAULTS:

                if lpd[0] == default:
                    self.frequency.value = lpd[1]
                    self.angle.value = lpd[2]

                    if lpd[3] in LPI.spot_xml:
                        self.spot = lpd[3]
                    else:
                        self.spot = "ellipse"

                    break

    def fromxml(self, xml: ET.Element) -> bool:

        if xml.tag != "LPI":
            return False

        if (color := xml.get("Color")) is not None:
            self.color = color

        if (freq := xml.get("Frequency")) is not None:
            self.frequency.value = int(freq)

        if (angle := xml.get("Angle")) is not None:
            self.angle.value = float(angle)

        if (spot := xml.get("SpotFunction")) is not None:

            for human, code in LPI.spot_xml.items():
                if code == spot:
                    self.spot = human
                    break

        return True

    def toxml(self) -> Union[ET.Element, None]:
        xml = ET.Element("LPI")

        if not self.color:
            return None

        xml.attrib["Color"] = self.color
        xml.attrib["Frequency"] = self.frequency.toxmlstr(True)
        xml.attrib["Angle"] = self.angle.toxmlstr(True)
        xml.attrib["SpotFunction"] = LPI.spot_xml[self.spot]

        return xml


class PrinterSettings(PyScribusElement):
    def __init__(self):
        super().__init__()

        self.mirror_pages = {"horizontal": False, "vertical": False}

        self.marks = {
            "crop": False,
            "bleed": False,
            "registration": False,
            "color": False,
        }

        firstUse = "1"

        # Print to file
        toFile = "0"
        useAltPrintCommand = "0"
        outputSeparations = "0"

        # Whether to use spot colors
        useSpotColors = "0"
        useColor = "0"

        # Use ICC color profiles
        useICC = "0"
        # Whether to have grey component replacement
        doGCR = "0"

        doClip = "0"
        setDevParam = "0"

        includePDFMarks = "0"

        # The postscrip level
        PSLevel = "0"

        # Which printer description language
        PDLanguage = "0"
        markLength = "7.185302734375"
        markOffset = "0"

        # --- Bleeds --------------------------------------------

        self.bleeds = collections.OrderedDict()

        self.bleeds["top"] = (dimensions.Dim(0),)
        self.bleeds["left"] = (dimensions.Dim(0),)
        self.bleeds["right"] = (dimensions.Dim(0),)
        self.bleeds["bottom"] = (dimensions.Dim(0),)
        self.bleeds["document"] = False

        # -------------------------------------------------------

        printer = ""
        filename = ""
        separationName = ""
        printerCommand = ""

    # --- PyScribus standard methods ----------------------------

    def fromxml(self, xml: ET.Element) -> bool:

        if xml.tag != "Printer":
            return False

        # --- Page mirroring ------------------------------------

        if (mrh := xml.get("mirrorH")) is not None:
            self.mirror_pages["horizontal"] = num_to_bool(mrh)

        if (mrv := xml.get("mirrorV")) is not None:
            self.mirror_pages["vertical"] = num_to_bool(mrv)

        # --- Marks ---------------------------------------------

        for case in ["crop", "bleed", "registration", "color"]:
            att_name = f"{case}Marks"

            if (att := xml.get(att_name)) is not None:
                self.marks[case] = num_to_bool(att)

        # --- Printer bleeds ------------------------------------

        for case in ["top", "left", "right", "bottom"]:
            att_name = "Bleed{}".format(case.capitalize())

            if (att := xml.get(att_name)) is not None:
                self.bleeds[case][0].value = float(att)

        if (udb := xml.get("useDocBleeds")) is not None:
            self.bleeds["document"] = num_to_bool(udb)

        # -------------------------------------------------------

        return True

    def toxml(self) -> ET.Element:
        xml = ET.Element("Printer")

        # TODO

        for bleed_config, bleed_value in self.bleeds.items():

            if bleed_config != "document":
                att_name = "Bleed{}".format(bleed_config.capitalize())
                xml.attrib[att_name] = bleed_value[0].toxmlstr(True)

        xml.attrib["mirrorH"] = bool_to_num(self.mirror_pages["horizontal"])
        xml.attrib["mirrorV"] = bool_to_num(self.mirror_pages["vertical"])

        xml.attrib["useDocBleeds"] = bool_to_num(self.bleeds["document"])

        for key, value in self.marks.items():
            att_name = f"{key}Marks"

            xml.attrib[att_name] = bool_to_num(value)

        return xml


# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus

Scribus file. Document UI settings classes.
"""

# Imports ===============================================================#

# Standard library ---------------------------------------------

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

import copy

from typing import Union, Any, NoReturn, Optional, Literal
from collections import OrderedDict

# PyScribus model ----------------------------------------------

from pyscribus.model.sla import SLA
from pyscribus.model.common.math import PicaConverter

# PyScribus file -----------------------------------------------

import pyscribus.file as pyf
import pyscribus.file.exceptions as pyfe

# Global variables / Annotations ========================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

BoolOrNone = Union[bool, None]

RenderOrderItems = Literal["baseline", "grids", "guides", "margins", "objects"]

# Tool classes ==========================================================#


class Tool:
    """
    Abstract class for the default settings, document-wide, of a Scribus tool.
    """

    def __init__(self, model: SLA, converter: Optional[PicaConverter] = None):
        self.model = model
        self._tool_id = None
        self._converter = converter

    def _picas(self, value: Union[int, float]) -> float:
        if self._converter is None:
            return value

        return self._converter.picas(value)

    @pyfe.has_document
    def _get(self, key) -> Union[Any, None]:
        return self.model.document.tools[self._tool_id].get(key)

    @pyfe.has_document
    def _subget(self, key, subkey) -> Union[Any, None]:
        return self.model.document.tools[self._tool_id][key].get(subkey)

    @pyfe.has_document
    def _set(self, key: str, value: Any, dimension=False) -> NoReturn:
        """
        Set a tool preference named ``key`` to ``value``.


        :type key: str
        :param key: Key of the tool preference.
        :type value: Any
        :param value: Value to set to the tool preference ``key``.
        :type dimension: bool
        :param dimension: Set the value of a preference that is a
            ``pyscribus.model.dimension.Dim`` object?
        """

        if dimension:
            self.model.document.tools[self._tool_id][key].value = value
        else:
            self.model.document.tools[self._tool_id][key] = value

    @pyfe.has_document
    def _subset(
        self,
        key: str,
        subkey: str,
        value: Any,
        dimension=False
    ) -> NoReturn:
        """
        Set a tool preference named ``key``:``subkey`` to ``value``.


        :type key: str
        :param key: Key of the tool preference.
        :type subkey: str
        :param subkey: Sub-key of the tool preference.
        :type value: Any
        :param value: Value to set to the tool preference ``key``.
        :type dimension: bool
        :param dimension: Set the value of a preference that is a
            ``pyscribus.model.dimension.Dim`` object?
        """

        if dimension:
            self.model.document.tools[self.tool_id][key][subkey].value = value
        else:
            self.model.document.tools[self.tool_id][key][subkey] = value


class TextTool(Tool):
    """
    Default settings, document-wide, of the Scribus tool “Text”.
    """

    def __init__(self, model: SLA, converter: Optional[PicaConverter] = None):
        Tool.__init__(self, model, converter)
        self._tool_id = "text"

    # Text size ------------------------------------------------

    @property
    def size(self) -> float:
        """Text size (in points)."""
        return self._get("size").value

    @size.setter
    def size(self, value: float) -> NoReturn:
        """Text size (setter)."""
        self._set("size", value, True)

    # Columns gap size -----------------------------------------

    @property
    def columns_gap(self) -> int:
        """Gap between columns size (in picas)"""
        return self._get("columns_gap")

    @columns_gap.setter
    def columns_gap(self, value: float) -> NoReturn:
        """Gap between columns size (setter)"""
        self._set("columns_gap", self._picas(value), True)

    # Font -----------------------------------------------------

    # TODO Font
    # I don't know how to represent/give access to font for now.
    # Maybe use an interface around fonts in common with file.document.Fonts ?

    # Columns --------------------------------------------------

    @property
    def columns(self) -> int:
        """Default column number."""
        return self._get("columns")

    @columns.setter
    def columns(self, value: float) -> NoReturn:
        """Default column number (setter)"""
        self._set("columns", value)


class ShapeTool(Tool):
    """
    Default settings, document-wide, of the Scribus tool “Shape”.
    """

    def __init__(self, model: SLA, converter: Optional[PicaConverter] = None):
        Tool.__init__(self, model, converter)
        self._tool_id = "shape"

    # ----------------------------------------------------------

    def __set_color(self, subtool: str, value: str):
        # TODO : Maybe use a Color object instead of a string
        self._subset(subtool, "color", value)

    def __get_color(self, subtool: str) -> str:
        return self._subget(subtool, "color")

    def __set_shade(self, subtool: str, value: int):
        # TODO : Maybe use a Color object instead of a string
        self._subset(subtool, "shade", value, True)

    def __get_shade(self, subtool: str) -> int:
        return self._subget(subtool, "shade")

    # Pen ------------------------------------------------------

    @property
    def pen_color(self) -> str:
        """Color used by the shape tool pen."""
        return self.__get_color("pen")

    @pen_color.setter
    def pen_color(self, value: str) -> NoReturn:
        """Color used by the shape tool pen (setter)."""
        # TODO : Maybe use a Color object instead of a string
        self.__set_color("pen", value)

    @property
    def pen_shade(self) -> int:
        """Opacity the shape tool pen (integer percentage)."""
        return self.__get_shade("pen")

    @pen_shade.setter
    def pen_shade(self, value: int) -> NoReturn:
        """Opacity the shape tool pen (setter)."""
        self.__set_shade("pen", value)

    # Brush ----------------------------------------------------

    @property
    def brush_color(self) -> str:
        """Color used by the shape tool brush."""
        return self.__get_color("brush")

    @brush_color.setter
    def brush_color(self, value: str) -> NoReturn:
        """Color used by the shape tool brush (setter)."""
        # TODO : Maybe use a Color object instead of a string
        self.__set_color("brush", value)

    @property
    def brush_shade(self) -> int:
        """Opacity the shape tool brush (integer percentage)."""
        return self.__get_shade("brush")

    @brush_shade.setter
    def brush_shade(self, value: int) -> NoReturn:
        """Opacity the shape tool brush (setter)."""
        self.__set_shade("brush", value)

    # Width ----------------------------------------------------

    @property
    def width(self) -> float:
        """Width of the shape (in picas)."""
        return self._get("width").value

    @width.setter
    def width(self, value: float) -> NoReturn:
        """Width of the shape (setter)."""
        self._set("width", self._picas(value), True)

    # Style ----------------------------------------------------

    @property
    def style(self) -> int:
        """Style of the shape (from 0 to 37)."""
        return self._get("style")

    @style.setter
    def style(self, value: int) -> NoReturn:
        """Style of the shape (setter)."""
        value = int(value)

        if value < 0 or value > 37:
            raise ValueError(
                f"Shape style value must range from 0 to 37, got {value}."
            )

        self._set("style", value)


class Tools:
    """
    Interface for document's default tools settings for Scribus.

    :Example:

    .. code:: python

       from pyscribus.file import ScribusFile

       doc = ScribusFile("1.5.5", "example.sla")

       scribus_ui_tools = doc.ui.tools

       scribus_ui_tools.text.size = 12
       scribus_ui_tools.shape.pen_shade = 50

    """

    def __init__(self, model: SLA, converter: Optional[PicaConverter] = None):
        self.model = model
        self.converter = converter

    @property
    def text(self) -> TextTool:
        """
        Default settings, document-wide, of the Scribus tool “Text”.

        .. seealso:: :class:`pyscribus.file.ui.TextTool`
        """
        return TextTool(self.model, self.converter)

    @property
    def shape(self) -> ShapeTool:
        """
        Default settings, document-wide, of the Scribus tool “Shape”.

        .. seealso:: :class:`pyscribus.file.ui.ShapeTool`
        """
        return ShapeTool(self.model, self.converter)

    # TODO Image tool
    # TODO Polygon tool
    # TODO Line tool
    # TODO Calligraphic pen tool
    # TODO Arc tool
    # TODO Spiral tool


class UIGrid:
    """
    Abstract class for interfaces around UI grids settings.
    """

    def __init__(self, model: SLA, converter: Optional[PicaConverter] = None):
        self.model = model
        self.grid_name = None
        self._converter = converter

    def _picas(self, value: Union[int, float]) -> float:
        if self._converter is None:
            return value

        return self._converter.picas(value)

    @pyfe.has_document
    def _get_grid_show(self) -> bool:
        return self.model.document.ui_show.get(self.grid_name)

    @pyfe.has_document
    def _set_grid_show(self, value) -> NoReturn:
        self.model.document.ui_show[self.grid_name] = bool(value)

    @pyfe.has_document
    def _set_grid_setting(
        self,
        key: str,
        value: Any,
        grid_name: Optional[str] = None,
        dimension: bool = False
    ) -> NoReturn:
        grid = None

        if grid_name is None:
            grid = self.grid_name
        else:
            grid = grid_name

        if dimension:
            self.model.document.grids[grid][key].value = value
        else:
            self.model.document.grids[grid][key] = value

    @pyfe.has_document
    def _get_grid_setting(
        self, attrib: str, grid_name: Optional[str] = None
    ) -> Any:
        if grid_name is None:
            return self.model.document.grids[self.grid_name].get(attrib)

        return self.model.document.grids[grid_name].get(attrib)

    @property
    def show(self) -> bool:
        """Show this grid?"""
        return self._get_grid_show()

    @show.setter
    def show(self, do_show: bool) -> NoReturn:
        self._set_grid_show(do_show)


class Baseline(UIGrid):
    """
    Interface to UI baseline settings.
    """

    def __init__(self, model: SLA):
        UIGrid.__init__(self, model)
        self.grid_name = "baseline"

    # Baseline color -----------------------------------------------------

    @property
    def color(self):
        """Base line color showed in Scribus UI."""
        return self._get_grid_setting("color")

    # Baseline step ------------------------------------------------------

    @property
    def step(self):
        """Baseline step (in points)."""
        return self._get_grid_setting("spacing")

    @step.setter
    def step(self, value: float) -> NoReturn:
        """Baseline step (setter)."""
        self._set_grid_setting("spacing", value, dimension=True)

    # Baseline offset ----------------------------------------------------

    @property
    def offset(self):
        """Baseline offset (in points)."""
        return self._get_grid_setting("offset")

    @offset.setter
    def offset(self, value: float) -> NoReturn:
        """Baseline offset (setter)."""
        self._set_grid_setting("offset", value, dimension=True)


class Grids(UIGrid):
    """
    Interface to UI grid settings.
    """

    def __init__(self, model: SLA):
        UIGrid.__init__(self, model)
        self.grid_name = "grid"

    # Minor grid ---------------------------------------------------------

    @property
    def minor_color(self):
        """Minor grid: Color showed in Scribus UI."""
        return self._get_grid_setting("color", grid_name="minor")

    @property
    def minor_step(self):
        """Minor grid: step."""
        return self._get_grid_setting("spacing", grid_name="minor")

    @minor_step.setter
    def minor_step(self, value: float) -> NoReturn:
        self._set_grid_setting("spacing", self._picas(value), "minor", True)

    # Major grid ---------------------------------------------------------

    @property
    def major_color(self):
        """Major grid: Color showed in Scribus UI."""
        return self._get_grid_setting("color", grid_name="major")

    @property
    def major_step(self):
        """Major grid: step."""
        return self._get_grid_setting("spacing", grid_name="major")

    @major_step.setter
    def major_step(self, value: float) -> NoReturn:
        self._set_grid_setting("spacing", self._picas(value), "major", True)


class ShowOrder:
    items = [
        "baseline",
        "grids",
        "guides",
        "margins",
        "objects",
    ]

    def __init__(self, model: SLA):
        self.model = model

        self.__order = OrderedDict(
            item_1=None,
            item_2=None,
            item_3=None,
            item_4=None,
            item_5=None,
        )

        self.__from_model()

    @property
    @pyfe.has_document
    def __model_stack(self):
        return self.model.document.ui_show["stack"]

    @pyfe.has_document
    def __from_model(self):
        for model_index, file_key in [
            [0, "item_1"],
            [1, "item_2"],
            [2, "item_3"],
            [3, "item_4"],
            [4, "item_5"],
        ]:
            self.__order[file_key] = self.__model_stack[model_index]

    @pyfe.has_document
    def __iter__(self):
        """
        Special method to make RenderOrder class iterable.
        """
        self.__apply_changes()

        for item in self.__model_stack:
            yield item

    @pyfe.has_document
    def __apply_changes(self):
        """
        Apply the render order of this interface to the model datas.
        """
        self.__model_stack["stack"] = list(self.__order.values())

    def move(self, ui_item: RenderOrderItems, new_order: int) -> bool:
        """
        Move the render order of the UI element ``ui_item`` to ``new_order``.

        :type ui_item: RenderOrderItems
        :param ui_item: UI element to change its render order
        :type new_order: int
        :param new_order: New render order of the UI element, ranging from 1
            to 5.
        :rtype: bool
        :returns: Success.
        """

        if ui_item not in ShowOrder.items:
            raise ValueError(
                "UI render item must be one of these values: {}.".format(
                    ", ".join([f"'{s}'" for s in ShowOrder.items])
                )
            )

        if new_order < 1 or new_order > 5:
            raise ValueError("UI render order must be between 1 to 5.")

        # Nothing to do

        if self.__order[f"order_{new_order}"] == ui_item:
            return True

        # Swap item and older item at new_order position

        old_order = None
        old_value = copy.deepcopy(self.__order[f"order_{new_order}"])

        for order_key, item_value in self.__order.items():
            if item_value == ui_item:
                old_order = order_key
                break

        self.__order[old_order] = old_value
        self.__order[f"order_{new_order}"] = ui_item

        self.__apply_changes()

        return True


# UI interface ==========================================================#


class UI:
    """
    Interface for document's UI settings for Scribus.

    :Example:

    .. code:: python

       from pyscribus.file import ScribusFile

       doc = ScribusFile("1.5.5", "example.sla", coding_unit="mm")

       scribus_ui = doc.ui

       # Make Scribus show invisible caracters and links between text frames
       scribus_ui.caracters = True
       scribus_ui.links = True

       # Some elements have more settings.
       scribus_ui.baseline.show = False
       scribus_ui.grids.show = True
       scribus_ui.grids.minor_step = 10
       scribus_ui.grids.major_step = 50

    """

    def __init__(self, file: pyf.ScribusFile):
        self.file = file
        self.model = file.model

    @pyfe.has_document
    def __get(self, key) -> Union[str, list[str], None]:
        return self.model.document.ui_show.get(key)

    # UI settings properties ---------------------------------------------

    # Scribus UI elements not strictely related to the document

    @property
    def rulers(self) -> BoolOrNone:
        """Show Scribus' rulers?"""
        return self.__get("rulers")

    @property
    def layer_hints(self) -> BoolOrNone:
        """Show Scribus' layer hints?"""
        return self.__get("layer_hints")

    # Default settings, document wide, of Scribus tools --------

    @property
    def tools(self) -> Tools:
        """
        Default settings, document-wide, of Scribus' tools.

        .. seealso:: :class:`pyscribus.file.ui.Tools`
        """
        return Tools(self.model, self.file.pica_converter)

    # Document grid, lines -------------------------------------

    @property
    def grids(self) -> Grids:
        """
        Document's minor and major grids.

        .. seealso:: :class:`pyscribus.file.ui.Grids`
        """
        return Grids(self.model)

    @property
    def guides(self) -> BoolOrNone:
        """Show document's guides?"""
        return self.__get("guides")

    @property
    def baseline(self) -> Baseline:
        """
        Document's base line.

        .. seealso:: :class:`pyscribus.file.ui.Baseline`
        """
        return Baseline(self.model)

    @property
    def margins(self) -> BoolOrNone:
        """Show document's margins?"""
        return self.__get("margins")

    # Page objects : images frames -----------------------------

    @property
    def images(self) -> BoolOrNone:
        """Show images of the images frames?"""
        return self.__get("images")

    # Page objects : text frames -------------------------------

    @property
    def links(self) -> BoolOrNone:
        """Show links between text frames?"""
        return self.__get("links")

    @property
    def text_columns(self) -> BoolOrNone:
        """Show the borders of text frames columns?"""
        return self.__get("colborders")

    @property
    def caracters(self) -> BoolOrNone:
        """Show invisible/control caracters in text frames?"""
        return self.__get("caracters")

    # Render order of everything in the UI ---------------------

    @property
    def show_order(self) -> ShowOrder:
        """
        Order of render of some Scribus UI elements (baseline, grids, guides,
        margins, objects).

        .. seealso:: :class:`pyscribus.file.ui.ShowOrder`
        """
        return ShowOrder(self.model)


# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus

Scribus file. Classes for page objects.
"""

# Imports ===============================================================#

# Standard library ---------------------------------------------

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

from typing import Union, Any

import fnmatch

# PyScribus model ----------------------------------------------

import pyscribus.model.dimensions as dimensions
import pyscribus.model.pageobjects as pageobjects

# PyScribus file -----------------------------------------------

import pyscribus.file as pyf
import pyscribus.file.exceptions as pyfe
import pyscribus.file.document as document

# Global variables / Annotations ========================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

StringOrNone = Union[str, None]

# Classes ===============================================================#

class PageObjects:
    """
    Interface for document's page objects.

    :type file: pyscribus.file.ScribusFile
    :param file: Scribus file wrapper

    :Example:

    .. code:: python

       # To avoid writing "scribus_file.pageobjects" again and again
       doc_objects = scribus_file.pageobjects

       # Filtering page objects ------------------------------------------

       templatable_text_frames = doc_objects.filter(
          object_type="text", templatable=True
       )

       on_page_two = doc_objects.filter(page_number=1)

       # Is an object on page 2 on page 1 ? No.
       is_on_page_one = doc_objects.on_page(on_page_two[0], 0)

       # Access page objects by loading order ----------------------------

       po_two = doc_objects[1]

       # To access all objects, use item property…

       all_objects = doc_objects.items

       # … or iterate on doc_objects:

       for object in doc_objects:
           print(object.name)

    """

    def __init__(self, file: pyf.ScribusFile):
        self.file = file
        self.model = file.model

    def __get_page(self, page_number: int, master: bool = False):
        """
        Get the page number ``page_number`` and the full page set (normal /
        master pages).

        :type page_number: int
        :param page_number: Page number of the page to copy.
        :type master: bool
        :param master: Use the master pages set instead of normal pages.
        :returns: Page number ``page_number`` and the full page set.
        """

        if page_number < 0:
            raise pyfe.PageNotFound("Page number asked inferior to 0.")

        if master:
            pages = document.Pages(self.model).masterpages
        else:
            pages = document.Pages(self.model).pages

        try:
            page = pages[page_number]
        except IndexError as not_found:
            raise pyfe.PageNotFound(
                f"There is no page number {page_number}"
            ) from not_found

        return page, pages

    def __on_page(self, page_object, page) -> bool:
        """
        Is the page object ``page_object`` on page ``page`` ?

        :rtype: bool
        """

        # We just use numeric values of the Dim objects
        def corner_values(corner: list[dimensions.Dim]) -> list[float]:
            return [dim.value for dim in corner]

        # Get the top-left and bottom-right corners coordinates of the page
        page_tl = corner_values(page.box.coords["top-left"])
        page_br = corner_values(page.box.coords["bottom-right"])
        # Get the top-left and bottom-right corners coordinates of the object
        object_tl = corner_values(page_object.box.coords["top-left"])
        object_br = corner_values(page_object.box.coords["bottom-right"])

        # Object in X of the page
        in_x = object_tl[0] >= page_tl[0] and object_br[0] <= page_br[0]
        # Object in Y of the page
        in_y = object_tl[1] >= page_tl[1] and object_br[1] <= page_br[1]
        # Object in the page
        on_page = in_x and in_y

        return on_page

    def __overlaps_page(
        self, page_object, page, strict: bool = False
    ) -> Union[bool, dict]:
        """
        Is the page object ``page_object`` overlaping on page ``page`` ?

        :param page_object: Page object
        :param page: Page.
        :type strict: bool
        :param strict: Filter out page object that are totally contained in
            the page.

        :rtype: Union[bool, dict]
        """

        # We just use numeric values of the Dim objects
        def corner_values(corner: list[dimensions.Dim]) -> list[float]:
            return [dim.value for dim in corner]

        obj = {
            corner_name: corner_values(corner_coords)
            for corner_name, corner_coords in page_object.box.coords.items()
        }

        page_tl = corner_values(page.box.coords["top-left"])
        page_br = corner_values(page.box.coords["bottom-right"])

        overlaping = {
            "top-left": False,
            "top-right": False,
            "bottom-left": False,
            "bottom-right": False,
        }

        for corner in ["top-left", "top-right", "bottom-left", "bottom-right"]:
            # Corner in X of the page
            in_x = (
                obj[corner][0] >= page_tl[0] and obj[corner][0] <= page_br[0]
            )
            # Corner in Y of the page
            in_y = (
                obj[corner][1] >= page_tl[1] and obj[corner][1] <= page_br[1]
            )
            # Corner in the page
            overlaping[corner] = in_x and in_y

        corner_overlaping = len(
            [status for status in overlaping.values() if status]
        )

        if corner_overlaping == 0:
            return False

        # If we want to know if the page object is overlaping on page but not
        # totally contained in that page, then having the four corners
        # overlaping should return False.
        if strict and corner_overlaping == 4:
            return False

        return overlaping

    def __is_model_object(self, page_object: Any) -> bool:
        """
        Is the object ``page_object`` an object from the list of the model
        page objects?

        :type page_object: Any
        :param page_object: A page object.
        :rtype: bool
        """

        for object_class in pageobjects.po_type_classes.values():
            if not isinstance(page_object, object_class):
                continue

            return True

        return False

    def append(self, page_object) -> bool:
        """
        Append the page object ``page_object``.

        .. warning:: This method only append page objects of the model
             classes. It does not do anything else and will change it's
             arguments and behaviour across versions.

        :rtype: bool
        :returns: Success.
        """

        if not self.__is_model_object(page_object):
            return False

        self.model.document.page_objects.append(page_object)

        return True

    def on_page(
        self,
        page_object,
        page_number: int,
        master: bool = False,
    ) -> bool:
        """
        Is the page object ``page_object`` on page number ``page_number`` ?

        .. note:: This methods detects objects strictely *contained* on pages
             — not objects overlaping in/out of pages. To detect overlaping
             objects, use ``overlaps_page`` method instead.

        :param page_object: Page object
        :type page_number: int
        :param page_number: Page number of the page to copy.
        :type master: bool
        :param master: Use the master pages set instead of normal pages.
        :rtype: bool
        :returns: ``True`` if the object is on the page.
        """

        # Get the relevant page
        page, _ = self.__get_page(page_number, master)

        return self.__on_page(page_object, page)

    def overlaps_page(
        self,
        page_object,
        page_number: int,
        master: bool = False,
        strict: bool = False,
        corners: bool = False,
    ) -> Union[bool, dict]:
        """
        Do the page object ``page_object`` overlaps on page number
        ``page_number`` ?

        .. warning:: This method checks only the non-rotated box of the page
            object.

        :param page_object: Page object
        :type page_number: int
        :param page_number: Page number.
        :type master: bool
        :param master: Use the master pages set instead of normal pages.
        :type strict: bool
        :param strict: Filter out page object that are totally contained in
            the page.
        :type corners: bool
        :param corners: Return the overlaping status of each corner of the page
            object. Returns a dictionary if there is overlaping corners, if
            not, returns ``False``.

        :rtype: Union[bool, dict]
        """
        page, _ = self.__get_page(page_number, master)

        overlaping = self.__overlaps_page(page_object, page, strict)

        if overlaping:
            if corners:
                return overlaping

            return True

        return False

    def __filter_name(self, lookup_set: list, pattern: StringOrNone) -> list:
        """
        Sub-filter for ``PageObjects.filter()``.

        Filter out page object not matching ``pattern``.

        Use fnmatch to match ``pattern`` like a file glob.

        :type lookup_set: list
        :param lookup_set: List of page objects to filter against.
        :type pattern: str
        :param page_number: Name pattern (glob) to match.
        :rtype: list
        """

        if pattern is None:
            return lookup_set

        return [
            page_object
            for page_object in lookup_set
            if fnmatch.fnmatchcase(page_object.name, pattern)
        ]

    def __filter_page(
        self,
        lookup_set: list,
        page_number: Union[int, None],
        master: Union[bool, None] = False,
    ) -> list:
        """
        Sub-filter for ``PageObjects.filter()``.

        Filter out page object not in the page number ``page_number``.

        :type lookup_set: list
        :param lookup_set: List of page objects to filter against.
        :type page_number: int
        :param page_number: Page number.
        :type master: bool
        :param master: Use the master pages set instead of normal pages.
        :rtype: list
        """

        if page_number is None:
            return lookup_set

        # If master is None, then the user did not specified master
        # in PageObjects.filter, so we can assume the user want regular pages.
        if master is None:
            master = False

        page, _ = self.__get_page(page_number, master)

        filtered_set = [
            page_object
            for page_object in lookup_set
            if self.__on_page(page_object, page)
        ]

        return filtered_set

    def __filter_page_overlap(
        self,
        lookup_set: list,
        page_number: Union[int, None],
        master: Union[bool, None] = False,
        strict: bool = False,
    ) -> list:
        """
        Sub-filter for ``PageObjects.filter()``.

        Filter out page object not overlaping on the page number ``page_number``.

        :type lookup_set: list
        :param lookup_set: List of page objects to filter against.
        :type page_number: int
        :param page_number: Page number.
        :type master: bool
        :param master: Use the master pages set instead of normal pages.
        :type strict: bool
        :param strict: Filter out page object that are totally contained in
            the page.
        :rtype: list
        """

        if page_number is None:
            return lookup_set

        # If master_page is None, then the user did not specified master_page
        # in PageObjects.filter, so we can assume the user want regular pages.
        if master is None:
            master = False

        page, _ = self.__get_page(page_number, master)

        filtered_set = [
            page_object
            for page_object in lookup_set
            if self.__overlaps_page(page_object, page, strict)
        ]

        return filtered_set

    def __filter_type(
        self, lookup_set: list, object_type: Union[str, bool, None]
    ) -> list:
        """
        Sub-filter for ``PageObjects.filter()``.

        Filter by object type.

        :type lookup_set: list
        :param lookup_set: List of page objects to filter against.
        :type object_type: Union[str, bool, None]
        :param object_type: Page object type to filter, or do not filter at
            all. See pageobjects.po_type_classes for valid values.
        :rtype: list
        """
        if object_type is None:
            return lookup_set

        if not object_type:
            return lookup_set

        if object_type not in pageobjects.po_type_classes:
            raise ValueError(
                f"Wrong object type when filtering page objects: {object_type}"
            )

        filtered_set = [
            page_object
            for page_object in lookup_set
            if isinstance(
                page_object, pageobjects.po_type_classes[object_type]
            )
        ]

        return filtered_set

    def __filter_templatable(
        self, lookup_set: list, templatable: Union[bool, None] = False
    ) -> list:
        """
        Sub-filter for ``PageObjects.filter()``.

        Filter out non-templatable page objects.

        :type lookup_set: list
        :param lookup_set: List of page objects to filter against.
        :type templatable: Union[bool, None]
        :param templatable: Only return templatable page objects.
        :rtype: list
        """
        if templatable is None:
            return lookup_set

        if not templatable:
            return lookup_set

        if not self.file.templating["active"]:
            return []

        filtered_set = []

        for object_item in lookup_set:
            # If the page object is a text frame with templatable
            # stories, we add these templatable stories

            if isinstance(object_item, pageobjects.TextObject):
                po_templatable_stories = object_item.templatable()

                if po_templatable_stories:
                    filtered_set.extend(po_templatable_stories)

            else:
                # TODO If this page object is another type of page
                # object, we look its properties and find if it
                # is templatable through sla.SLA.templating settings

                if object_item.templatable():
                    filtered_set.append(object_item)

        return filtered_set

    @pyfe.has_document
    def filter(self, **kwargs) -> list:
        """
        Return document page objects.

        :rtype: list
        :returns: Filtered document page objects.

        .. seealso:: ``pyscribus.model.pageobjects.po_type_classes`` for valid
            values expected by ``object_type`` filter.

        +---------------------+-------------------------------------------+------+
        | Kwarg key           | Filter on                                 | Type |
        +=====================+===========================================+======+
        | object_type         | Object type                               | str  |
        +---------------------+-------------------------------------------+------+
        | name                | Name of the page object. Glob-like and    | str  |
        |                     | case sensitive.                           |      |
        +---------------------+-------------------------------------------+------+
        | templatable         | Templatable objects                       | bool |
        +---------------------+-------------------------------------------+------+
        | page                | Objects in page X                         | int  |
        +---------------------+-------------------------------------------+------+
        | page_overlap        | Objects overlaping in page X.             | int  |
        |                     |                                           |      |
        |                     | Objects totally contained in page X are   |      |
        |                     | also returned.                            |      |
        +---------------------+-------------------------------------------+------+
        | page_overlap_strict | Objects overlaping in page X.             | int  |
        |                     |                                           |      |
        |                     | Objects totally contained in page X are   |      |
        |                     | NOT returned.                             |      |
        +---------------------+-------------------------------------------+------+
        | master_page         | When using the filters ``page`` and       | bool |
        |                     | ``page_overlap``, filter on master pages  |      |
        +---------------------+-------------------------------------------+------+

        :Example:

        .. code:: python

           # To avoid writing "scribus_file.pageobjects" again and again
           doc_objects = scribus_file.pageobjects

           templatable_text_frames = doc_objects.filter(
              object_type="text", templatable=True
           )

           on_page_two = doc_objects.filter(page=1)

           # Returns all the objects named after daleks, so we can EXTERMINATE
           # them instead of us humans.
           daleks = doc_objects.filter(name="Dalek *")

        """

        # Filter on page object name -------------------------------------

        filtered_set = self.__filter_name(self.items, kwargs.get("name"))

        # Filter on page object type -------------------------------------

        filtered_set = self.__filter_type(
            filtered_set, kwargs.get("object_type")
        )

        # Filter on page -------------------------------------------------

        filtered_set = self.__filter_page(
            filtered_set, kwargs.get("page"), kwargs.get("master_page")
        )

        # Filter on overlaping pages -------------------------------------

        filtered_set = self.__filter_page_overlap(
            filtered_set,
            kwargs.get("page_overlap"),
            kwargs.get("master_page"),
            False,
        )

        # Filter on overlaping pages (strict) ----------------------------

        filtered_set = self.__filter_page_overlap(
            filtered_set,
            kwargs.get("page_overlap_strict"),
            kwargs.get("master_page"),
            True,
        )

        # Filter off templatable page objects ----------------------------

        filtered_set = self.__filter_templatable(
            filtered_set, kwargs.get("templatable")
        )

        # ----------------------------------------------------------------

        return filtered_set

    @property
    @pyfe.has_document
    def items(self) -> list:
        """List of all page objects."""
        return self.model.document.page_objects

    def __iter__(self):
        for page_object in self.model.document.page_objects:
            yield page_object

    def __getitem__(self, item: int):
        return self.model.document.page_objects[item]


# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Mathematics related values, functions and enumerations.
"""

# Imports ===============================================================#

import enum
import math

from typing import Union, Literal, Optional

# Global variables / annotations ========================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# 1 pica point = 25,4/72 milimeters
PICA_TO_MM = 25.4 / 72

# 1 inch = 25,4 milimeters
INCH_TO_MM = 25.4

# 1 cicero = 1.066 picas
CICERO_TO_PICA = 1.066

ConvertableToPicas = Literal["mm", "in", "c"]

# Classes ===============================================================#

class FloatEnum(enum.Enum):
    """
    Enum class usable with float() function / method.
    """

    def __float__(self):
        if isinstance(self.value, float):
            return self.value

        return float(self.value)

# Fonctions =============================================================#


def truncate(number, digits) -> float:
    """
    Truncate value `number` to X `digits` digits.

    From Stack Overflow.

    <https://stackoverflow.com/questions/8595973/truncate-to-three-decimals-in-python>

    :type number: float
    :type digits: int
    :rtype: float
    """

    dec_number = len(str(number).split('.')[1])

    if dec_number <= digits:
        return number

    stepper = 10.0 ** digits

    return math.trunc(stepper * number) / stepper


def necessary_float(f: float) -> Union[float, int]:
    """
    Return **integer** if float f has no decimals, else returns **float**.

    :type f: float
    :param f: Float value
    :rtype: int,float
    :returns: Integer if float f has no decimals, else float.
    """

    if float(f) == int(f):
        return int(f)

    return float(f)


def pica(value: float, unit: ConvertableToPicas) -> float:
    """
    Return value ``value`` in original unit ``unit`` in pica points.

    :rtype: float
    """

    value = float(value)

    if unit == "in":
        return inch(value)
    if unit == "mm":
        return mm(value)
    if unit == "c":
        return cicero(value)

    raise ValueError(f"Unit {unit} not known by pica conversion function.")


def cicero(ciceros: Union[int, float]) -> float:
    """
    Returns ciceros in pica points.

    :type inches: int,float
    :param ciceros: Ciceros
    :rtype: float
    """
    if isinstance(ciceros, int):
        ciceros = float(ciceros)

    return ciceros / CICERO_TO_PICA


def inch(inches: Union[int, float]) -> float:
    """
    Returns inches in pica points.

    :type inches: int,float
    :param inches: Inches
    :rtype: float
    """
    if isinstance(inches, int):
        inches = float(inches)

    return mm(inches / INCH_TO_MM)


def mm(milimeters: Union[int, float]) -> float:
    """
    Returns milimeters in pica points.

    :type milimeters: int,float
    :param milimeters: Milimeters
    :rtype: float
    """
    if isinstance(milimeters, int):
        milimeters = float(milimeters)

    return milimeters / PICA_TO_MM


# Converter classes =====================================================#

class PicaConverter:
    """
    Converter into picas class.

    Define a original units, output picas.
    """

    convertable_units = {
        "millimeters": "mm",
        "inches": "in",
        "ciceros": "c",
    }

    def __init__(self, original_unit: Optional[str] = None):
        self.original_unit = original_unit

    def set_unit(self, original_unit: Optional[str] = None) -> bool:
        for long, short in PicaConverter.convertable_units.items():

            if original_unit == short:
                self.original_unit = long
                break

            if original_unit == long:
                self.original_unit = long
                break

        if self.original_unit is None:
            self.original_unit = None
            return None

        return self.original_unit

    def picas(self, value: Union[int, float]) -> float:
        """
        Return value ``value`` in pica points.

        :rtype: float
        """

        if self.original_unit is None:
            return value

        return pica(value, PicaConverter.convertable_units[self.original_unit])

# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus

Scribus file. Exceptions classes and decorators related to them.
"""

# Imports ===============================================================#

import functools

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#


class NoDocumentError(Exception):
    """Raised when a model has no document."""


class NoModelError(Exception):
    """Raised when a ScribusFile has no underlying model."""


class PageNotFound(Exception):
    """Raised when the user asks for a page that can't be found."""


class LayerNotFound(Exception):
    """Raised when the user asks for a layer that can't be found."""


# Decorators ============================================================#


def has_model(func):
    """
    Decorator. Raise NoModelError if ScribusFile has no model.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kw):
        if self.model is None:
            raise NoModelError(
                "No underlying model found in this ScribusFile."
            )

        return func(self, *args, **kw)

    return wrapper


def has_document(func):
    """
    Decorator. Raise NoDocumentError if ScribusFile has no model.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kw):
        if self.model is None:
            raise NoModelError(
                "No underlying model found in this ScribusFile."
            )

        if self.model.document is None:
            raise NoDocumentError()

        return func(self, *args, **kw)

    return wrapper


# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus

High-level interface to Scribus file module.
"""

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

import re
import functools

from typing import Union, Optional, NoReturn, Literal
from pathlib import Path

import lxml.etree as ET

import pyscribus.file.exceptions as pyfe

from pyscribus.file.ui import UI
from pyscribus.file.colors import Colors
from pyscribus.file.stories import Stories
from pyscribus.file.document import Metadatas, Pages
from pyscribus.file.pageobjects import PageObjects
from pyscribus.file.hyphenation import Hyphenation

# TODO Implement those
# from pyscribus.file.document import Fonts, Layers

from pyscribus.model.sla import SLA
from pyscribus.model.common.math import PicaConverter

# Global variables / annotations ========================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

StringOrPath = Union[str, Path]

UnitString = Literal[
    "points",
    "pt",
    "millimeters",
    "mm",
    "inches",
    "in",
    "picas",
    "p",
    "centimeters",
    "cm",
    "ciceros",
    "c",
]

CodingUnitString = Literal[
    "millimeters",
    "mm",
    "inches",
    "in",
    "ciceros",
    "c",
]

# Classes ===============================================================#


class ScribusFile:
    """
    High-level interface to Scribus file.

    :type filepath: str
    :param filepath: Scribus file path
    :type version: str
    :param version: Scribus version (ex. '1.5.1')
    :type kwargs: dict
    :param kwargs: kwargs

    **Kwargs options :**

    +-----------------------+----------------------------+---------------+
    | Kwarg key             | Use                        | Default value |
    +=======================+============================+===============+
    | templating            | Use templating ?           | ``False``     |
    +-----------------------+----------------------------+---------------+
    | templatingInsensitive | Should in-text templating  | ``False``     |
    |                       | be case insensitive ?      |               |
    +-----------------------+----------------------------+---------------+
    | templatingPattern     | compiled regex to find     | ``\^%\w+%$``  |
    |                       | templated elements         |               |
    |                       | (ex: %TITLE%)              |               |
    +-----------------------+----------------------------+---------------+
    | units                 | Units used in the document | ``"points"``  |
    +-----------------------+----------------------------+---------------+
    """

    def __init__(
        self, version: str, load: Optional[StringOrPath] = None, **kwargs
    ):
        # “Low-level” representation of Scribus format
        self.model = None

        self.settings = {
            "coding_unit": None,
        }

        self.pica_converter = PicaConverter()

        # Templating --------------------------------------------------------

        self.templating = {
            "active": False,
            # In text templating sequences are like %Title%
            # "intext-pattern": re.compile("^%\w+%$"),
            "intext-pattern": re.compile("(%\w+%)+"),
            # %Title% = %title% = %TITLE% ?
            "intext-insensitive": False,
            # Page object attribute templating names are like %Title%
            "attribute-pattern": re.compile("(%\w+%)+"),
            # dict for evaluating an item attribute value as python boolean
            "attribute-eval-bool-keywords": {"true": "True", "false": "False"},
            # Insensitive item attribute value evaluation as boolean ?
            "attribute-eval-bool-insensive": False,
        }

        for argname, argvalue in kwargs.items():
            if argname == "templating" and argvalue:
                self.templating["active"] = True

            if argname == "templatingInsensitive" and argvalue:
                self.templating["intext-insensitive"] = True

            if argname == "templatingPattern":
                try:
                    pattern = re.compile(argvalue)
                    self.templating["intext-pattern"] = pattern

                except TypeError as not_re_pattern:
                    raise TypeError(
                        "templating-pattern must be a re pattern string."
                    ) from not_re_pattern

        # -------------------------------------------------------------------

        if load is None:
            self.model = SLA(filepath=None, version=version)
        else:
            self.load(load, version)

        # -------------------------------------------------------------------

        ensure_model = kwargs.get("ensure_model")

        if ensure_model is None:
            ensure_model = True
        else:
            ensure_model = bool(ensure_model)

        if self.model is None and ensure_model:
            raise pyfe.NoModelError(
                "No underlying model found in this ScribusFile."
            )

        # if self.model:
        # self.model.templating = self.templating

        # Handling other arguments in kwargs --------------------------------

        if kwargs:
            self._quick_setup(kwargs)

    # Initialisation methods ---------------------------------------------

    def _quick_setup(self, settings: dict) -> NoReturn:
        for setting_name, setting_value in settings.items():
            # Document units ----------------------------------------------

            if setting_name == "unit":
                self.set_units(setting_value)

            # Settings unit -----------------------------------------------

            if setting_name == "coding_unit":
                self.set_coding_unit(setting_value)

    @pyfe.has_model
    def fromdefault(self) -> bool:
        return self.model.fromdefault()

    def load(self, sla_path: StringOrPath, version: str) -> Union[SLA, bool]:
        """
        Load a SLA file from its path ``sla_path``.

        :type sla_path: StringOrPath
        :param sla_path: Path of the SLA file.
        :type version: str
        :param version: Scribus version
        :rtype: Union[SLA, bool]
        :returns: SLA model or ``False``.
        """

        model = SLA(filepath=None, version=version)
        loaded = model.parse(sla_path)

        if loaded:
            self.model = model

            return model

        return False

    # --------------------------------------------------------------------

    @pyfe.has_model
    def set_units(self, unit: UnitString) -> bool:
        """
        Set the unit used in the document.

        +--------------+------------------------------+
        | Unit         | Short form | Long form       |
        +==============+============+=================+
        | Centimeters  | ``cm``     | ``centimeters`` |
        +--------------+------------+-----------------+
        | Ciceros      | ``c``      | ``ciceros``     |
        +--------------+------------+-----------------+
        | Inches       | ``in``     | ``inches``      |
        +--------------+------------+-----------------+
        | Millimeters  | ``mm``     | ``millimeters`` |
        +--------------+------------+-----------------+
        | Picas        | ``p``      | ``millimeters`` |
        +--------------+------------+-----------------+
        | Points       | ``pt``     | ``points``      |
        +--------------+------------+-----------------+

        :rtype: bool
        :returns: Success.
        """
        for model, abrev in [
            ["points", "pt"],
            ["millimeters", "mm"],
            ["inches", "in"],
            ["picas", "p"],
            ["centimeters", "cm"],
            ["ciceros", "c"],
        ]:
            if unit == abrev:
                self.model.document.units = model
                return True

            if unit == model:
                self.model.document.units = model
                return True

        return False

    @pyfe.has_model
    def set_font(self, **kwargs) -> bool:
        """
        .. warning:: Not implemented.

        """
        # """
        # Définit la fonte pour le fichier, le style, l’élément de Story désigné.
        # """
        raise NotImplementedError()

    # XML properties -----------------------------------------------------

    @property
    def xml_loaded(self) -> Union[None, ET.Element]:
        """
        XML node of the source XML file, if it exists.
        """
        return self.model.xml_copy

    @property
    def xml_output(self) -> ET.Element:
        """
        XML node of the document produced by the underlying model.
        """
        return self.model.toxml()

    # Properties ---------------------------------------------------------

    @property
    @pyfe.has_model
    def colors(self):
        """
        Interface to the colors of the document.

        .. seealso:: :class:`pyscribus.file.colors.Colors`
        """
        return Colors(self.model)

    @property
    @pyfe.has_model
    def fonts(self):
        """
        .. warning:: Not implemented.

        Interface to the fonts of the document.
        """
        raise NotImplementedError()

    @property
    @pyfe.has_model
    def hyphenation(self):
        """
        Interface to the hyphenation settings of the document.

        .. seealso:: :class:`pyscribus.file.hyphenation.Hyphenation`
        """
        return Hyphenation(self.model)

    @property
    @pyfe.has_model
    def layers(self):
        """
        .. warning:: Not implemented.

        Interface to the layers of the document.
        """
        raise NotImplementedError()

    @property
    @pyfe.has_model
    def metadatas(self):
        """
        Interface to the metadatas of the document.

        .. seealso:: :class:`pyscribus.file.document.Metadatas`
        """
        return Metadatas(self.model)

    @property
    @pyfe.has_model
    def pageobjects(self) -> list:
        """
        Interface to the page objects of the document.

        .. seealso:: :class:`pyscribus.file.document.PageObjects`
        """
        # return self.model.document.pageobjects(object_type, templatable)
        return PageObjects(self)

    @property
    @pyfe.has_model
    def pages(self):
        """
        Interface to the pages of the document.

        .. seealso:: :class:`pyscribus.file.document.Pages`
        """
        return Pages(self.model)

    @property
    @pyfe.has_model
    def ui(self):
        """
        Interface to the document's settings related to Scribus UI.

        .. seealso:: :class:`pyscribus.file.ui.UI`
        """
        return UI(self)

    # Properties related to stories --------------------------------------

    @property
    @pyfe.has_model
    def stories(self) -> list:
        """
        Interface to all the stories of the document.

        .. seealso:: :class:`pyscribus.file.stories.Stories`
        """
        return Stories(self)

    @property
    @pyfe.has_model
    def templatable_stories(self) -> list:
        """
        Interface to the templatable stories of the document.

        .. seealso:: :class:`pyscribus.file.stories.Stories`
        """
        return Stories(self, templatable=True)

    # PyScribus settings -------------------------------------------------

    def set_coding_unit(self, unit: CodingUnitString) -> NoReturn:
        """
        Set the unit used in programming.

        (You might want to use inches in a document configured for millimeters.)

        +--------------+------------------------------+
        | Unit         | Short form | Long form       |
        +==============+============+=================+
        | Ciceros      | ``c``      | ``ciceros``     |
        +--------------+------------+-----------------+
        | Inches       | ``in``     | ``inches``      |
        +--------------+------------+-----------------+
        | Millimeters  | ``mm``     | ``millimeters`` |
        +--------------+------------+-----------------+
        """
        self.settings["coding_unit"] = self.pica_converter.set_unit(unit)

    # --------------------------------------------------------------------

    @pyfe.has_model
    def save(self, filepath: StringOrPath) -> bool:
        """
        Save the file.

        :type filepath: StringOrPath
        :param filepath: SLA file path
        :rtype: boolean
        :returns: ``True`` if successfull
        """
        return self.model.save(filepath)


# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus

Scribus file. Stories wrapper.
"""

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

from typing import Union, Literal

import pyscribus.file as pyf
import pyscribus.file.exceptions as pyfe
import pyscribus.model.stories as model_stories

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

StoryFieldType = Literal["page_number", "page_count"]

# Classes ===============================================================#


class Stories:
    """
    Wrapper around Scribus stories.

    :type file: pyscribus.file.ScribusFile
    :param file: Scribus file wrapper
    :type templatable: bool
    :param templatable: Only load templatable stories

    :Example:

    .. code:: python

       from pyscribus.file import ScribusFile
       from pyscribus.file import Paragraph

       doc = ScribusFile("1.5.5", "example.sla")

       len(doc.stories)
       first_story = doc.stories[0]
       text = doc.stories.rawtext(first_story)

       for story in doc.stories.templatable:
           doc.stories.feed(story, data)

       # BELOW THIS LINE, EVERYTHING IS NOT IMPLEMENTED --------

    """

    def __init__(self, file: pyf.ScribusFile, templatable: bool = False):
        self.file = file
        self.model = file.model

        self.only_templatable = templatable

    def __len__(self) -> int:
        """

        :Example:

        .. code:: python

           from pyscribus.file import ScribusFile
           doc = ScribusFile("1.5.5", "example.sla")
           len(doc.stories)

        """
        return len(self.__stories())

    def __getitem__(self, item: int):
        """

        :Example:

        .. code:: python

           from pyscribus.file import ScribusFile
           doc = ScribusFile("1.5.5", "example.sla")
           first_story = doc.stories[0]

        """
        return self.__stories()[item]

    @property
    def items(self) -> list:
        """List of all stories."""
        return self.__stories()

    @property
    def templatable(self) -> list[model_stories.Story]:
        """
        Stories that contains templatable elements.

        :Example:

        .. code:: python

           from pyscribus.file import ScribusFile
           doc = ScribusFile("1.5.5", "example.sla")

           for story in doc.stories.templatable:
               doc.stories.feed(story, data)

        """
        return [
            story
            for story in self.__stories()
            if self.__story_is_templatable(story)
        ]

    def rawtext(self, story: model_stories.Story) -> str:
        """
        Returns a text string equivalent to **what Scribus story editor
        saves** as txt file.

        :Example:

        .. code:: python

           from pyscribus.file import ScribusFile
           doc = ScribusFile("1.5.5", "example.sla")

           story = doc.stories[1]
           text = doc.stories.rawtext(story)

        :rtype: string
        """
        text = ""

        for par in story.bypars():
            for element in par:
                # Text formatting is not saved
                if isinstance(element, model_stories.StoryFragment):
                    text += element.text
                if isinstance(element, model_stories.NonBreakingSpace):
                    text += " "
                # Line breaks are exported by Scribus... as spaces
                if isinstance(element, model_stories.StoryLineBreak):
                    text += " "
                # We dont care about StoryFrameBreak and StoryColumnBreak
                if isinstance(element, model_stories.StoryParagraphEnding):
                    text += "\n"

        return text

    def feed(
        self, story: model_stories.Story, datas: dict
    ) -> Union[bool, list[model_stories.StoryFragment]]:
        """
        Feed template-able datas into this story.

        :Example:

        .. code:: python

           from pyscribus.file import ScribusFile
           doc = ScribusFile("1.5.5", "example.sla")

           story = doc.stories[0]
           doc.stories.feed(story, data)

        :type story: pyscribus.model.stories.Story
        :type datas: dict
        :returns: List of modified elements in this story, or False
        :rtype: Union[bool, list]
        """

        elements = self.__story_templatable(story)

        if not elements:
            return False

        modified = []

        for element in elements:
            for template_key, template_value in datas.items():
                if template_key not in element.text:
                    continue

                new_value = None

                if isinstance(datas[template_key], list):
                    # We have %key% -> ["Value1", "Value2", …]
                    # So the first occurrence of %key% is "Value1", the
                    # second "Value2", etc.

                    if len(datas[template_key]):
                        new_value = datas[template_key].pop(0)
                else:
                    # We have %key% -> "Value"
                    new_value = template_value

                if new_value is None:
                    continue

                element.text = element.text.replace(template_key, new_value)

                modified.append(element)

        return modified

    def __story_templatable(self, story) -> list:
        """
        Return elements of Story sequence that are available for templating.

        :type story: pyscribus.model.stories.Story
        :rtype: list
        :returns: List of pyscribus.stories.StoryFragment
        """

        contents = []

        pattern = self.file.templating["intext-pattern"]

        for element in story.model_object.sequence:
            if isinstance(element, model_stories.StoryFragment):
                if pattern.findall(element.text) is not None:
                    contents.append(element)

        return contents

    def __story_is_templatable(self, story) -> bool:
        pattern = self.file.templating["intext-pattern"]

        for element in story.model_object.sequence:
            if isinstance(element, model_stories.StoryFragment):
                if pattern.findall(element.text) is not None:
                    return True

        return False

    @pyfe.has_document
    def __stories(self) -> list[model_stories.Story]:
        """
        Returns all stories in the document.

        :rtype: list
        :returns: List of stories
        """

        stories = []

        # Text frames stories --------------------------------------------

        filtered = [
            page_object
            for page_object in self.model.document.page_objects
            if page_object.have_stories and page_object.stories
        ]

        if filtered:
            for page_object in filtered:
                stories.extend(page_object.stories)

        # Table cells stories --------------------------------------------

        tables = [
            page_object
            for page_object in self.model.document.page_objects
            if page_object.ptype == "table"
        ]

        if tables:
            cells = []

            for page_object in tables:
                cells.extend(page_object.cells)

            for cell in cells:
                if cell.story is not None:
                    stories.append(cell.story)

        # Convert stories ------------------------------------------------

        # This converts stories from the model subpackage to stories
        # from the file subpackage.
        stories = [Story(model_object=story) for story in stories]

        # ----------------------------------------------------------------

        if self.only_templatable:
            return [
                story
                for story in stories
                if self.__story_is_templatable(story)
            ]

        return stories


class Story:
    """
    High-level wrapper around model's ``Story`` class.
    """

    def __init__(self, **kwargs):
        self.sequence = []
        self.model_object = None

        if (seq := kwargs.get("sequence")) is not None:
            self.sequence = seq

        if (mstory := kwargs.get("model_object")) is not None:
            self.model_object = mstory

    def append(self, item):
        """
        Append something to this story.

        .. warning:: Not implemented.

        """
        raise NotImplementedError()

    def __notes(self) -> list:
        """
        Returns the list of notes used in this story.

        .. warning:: Not implemented.

        """
        raise NotImplementedError()


class Paragraph:
    """
    Paragraph in a story.

    High-level wrapper around many elements that can make a paragraph in a
    Scribus story.

    (This class doesn't match any SLA element itself)
    """

    def __init__(self, **kwargs):
        self.sequence = []

        if (seq := kwargs.get("sequence")) is not None:
            self.sequence = seq

    def append(self, item):
        """
        .. warning:: Not implemented.

        """
        raise NotImplementedError()

    def len(self) -> int:
        """
        .. warning:: Not implemented.

        """
        raise NotImplementedError()

    def __notes(self) -> list:
        """
        .. warning:: Not implemented.

        """
        raise NotImplementedError()


class Notes:
    """
    Notes in a story or paragraph.
    """


class Fragment:
    def __init__(self, **kwargs):
        self.text = ""

        if (text := kwargs.get("text")) is not None:
            self.text = text

    def append(self, item):
        """
        .. warning:: Not implemented.

        """
        raise NotImplementedError()


class SpanText(Fragment):
    """
    Formatted fragment of text in a paragraph
    """


class Text(Fragment):
    """
    Fragment of text in a paragraph, without any formatting.
    """


class Field:
    """
    Special field in a Story paragraph
    """

    var_types = {
        "page_number": "pgno",
        "page_count": "pgco",
    }

    def __init__(self, field_type: StoryFieldType):
        self.model_object = None

        if field_type not in Field.var_types:
            raise ValueError("Unknown field name.")

        self.field_type = field_type

        self.model_object = {
            "pgno": model_stories.PageNumberVariable,
            "pgco": model_stories.PageCountVariable,
        }[Field.var_types[field_type]]()


# vim:set shiftwidth=4 softtabstop=4:
